#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
%prog chrC01 chr.sizes data

Read coverage histogram, similar to wiggle plot. Data contains all the track
data in the form of tab-delimited (x, y) lists.
"""

import os.path as op
import sys
import logging

import numpy as np

from jcvi.formats.sizes import Sizes
from jcvi.graphics.base import (
    plt,
    savefig,
    Rectangle,
    mb_formatter,
    mb_float_formatter,
    adjust_spines,
    get_map,
)
from jcvi.apps.base import OptionParser, glob


class XYtrack(object):
    def __init__(self, ax, datafile, color=None, ymax=40):
        self.ax = ax
        self.xy = []
        fp = open(datafile)
        for row in fp:
            atoms = row.split()
            self.xy.append([int(atoms[0]), float(atoms[1])])
        fp.close()

        self.x, self.y = zip(*self.xy)
        logging.debug(
            "File `{0}` imported (records={1}).".format(datafile, len(self.x))
        )
        self.color = color or "k"
        self.ymax = ymax

    @property
    def mapping(self):
        return dict(zip(self.x, self.y))

    def interpolate(self, maxsize, unit=10000):
        maxsize = int(maxsize)
        for pos in range(unit, maxsize + unit, unit):
            if pos in self.x:
                continue
            self.xy.append([pos, 0])
        self.xy.sort()
        self.x, self.y = zip(*self.xy)
        logging.debug("After interpolate: {0}".format(len(self.x)))

    def cap(self, ymax):
        self.xy = [[a, 0] if b > ymax else [a, b] for a, b in self.xy]
        self.x, self.y = zip(*self.xy)

    def draw(self):
        ax = self.ax
        color = self.color
        ax.plot(self.x, self.y, lw=0)
        ax.fill_between(self.x, self.y, color=color, lw=0)
        ax.set_ylim(0, self.ymax)
        ax.set_axis_off()

    def import_hlfile(self, hlfile, chr, unit=10000, diverge=("r", "g")):
        rr, gg = diverge
        fp = open(hlfile)
        imported = 0
        mapping = self.mapping
        for row in fp:
            if row.strip() == "":
                continue
            seqid, start, end, tag = row.split()
            if seqid != chr:
                continue
            start = int(start)
            end = int(end)
            if tag == "double":
                self.highlight(mapping, start, end, color=rr, unit=unit)
            else:
                self.highlight(mapping, start, end, color=gg, unit=unit)
            imported += 1
        logging.debug("Imported {0} regions from file `{1}`.".format(imported, hlfile))

    def highlight(self, mapping, start, end, color="r", unit=10000, zorder=10):
        ax = self.ax
        x = range(start, end + unit, unit)
        y = [mapping[z] for z in x]
        # Mask the highlight region so that they don't appear in background
        for a in self.xy:
            if start <= a[0] <= end:
                a[1] = 0
        self.x, self.y = zip(*self.xy)
        ax.plot(x, y, lw=0)
        ax.fill_between(x, y, color=color, lw=0, zorder=zorder)

    def vlines(self, xs, color="m"):
        for x in xs:
            self.ax.plot((x, x), (0, self.ymax), "-", color=color, lw=2)


class Coverage(object):
    def __init__(
        self,
        fig,
        root,
        canvas,
        chr,
        xlim,
        datadir,
        order=None,
        hlsuffix=None,
        palette=None,
        cap=50,
        gauge="bottom",
        plot_label=True,
        plot_chr_label=True,
        gauge_step=5000000,
        vlines=None,
        labels_dict={},
        diverge=("r", "g"),
    ):
        x, y, w, h = canvas
        p = 0.01
        root.add_patch(
            Rectangle(
                (x - p, y - p),
                w + 2 * p,
                h + 2 * p,
                lw=1,
                fill=False,
                ec="darkslategray",
                zorder=10,
            )
        )
        datafiles = glob(op.join(datadir, chr + "*"))

        if order:
            datafiles = [z for z in datafiles if z.split(".")[1] in order]
            datafiles.sort(key=lambda x: order.index(x.split(".")[1]))

        ntracks = len(datafiles)
        yinterval = h / ntracks
        yy = y + h

        if palette is None:
            # Get the palette
            set2 = get_map("Set2", "qualitative", ntracks).mpl_colors
        else:
            set2 = [palette] * ntracks

        if gauge == "top":
            gauge_ax = fig.add_axes([x, yy + p, w, 0.0001])
            adjust_spines(gauge_ax, ["top"])
            tpos = yy + 0.07
        elif gauge == "bottom":
            gauge_ax = fig.add_axes([x, y - p, w, 0.0001])
            adjust_spines(gauge_ax, ["bottom"])
            tpos = y - 0.07

        start, end = xlim
        if gauge:
            fs = gauge_step < 1000000
            setup_gauge_ax(gauge_ax, start, end, gauge_step, float_formatter=fs)

        if plot_chr_label:
            root.text(
                x + w / 2,
                tpos,
                chr,
                ha="center",
                va="center",
                color="darkslategray",
                size=16,
            )

        yys = []
        for label, datafile, c in zip(order, datafiles, set2):
            yy -= yinterval
            yys.append(yy)
            ax = fig.add_axes([x, yy, w, yinterval * 0.9])
            xy = XYtrack(ax, datafile, color=c)
            xy.interpolate(end)
            xy.cap(ymax=cap)
            if vlines:
                xy.vlines(vlines)
            if hlsuffix:
                hlfile = op.join(datadir, ".".join((label, hlsuffix)))
                xy.import_hlfile(hlfile, chr, diverge=diverge)
            if plot_label:
                label = labels_dict.get(label, label.capitalize())
                label = r"\textit{{{0}}}".format(label)
                root.text(x - 0.015, yy + yinterval / 2, label, ha="right", va="center")
            xy.draw()
            ax.set_xlim(*xlim)

        self.yys = yys


def setup_gauge_ax(gauge_ax, start, end, gauge_step, float_formatter=False):
    gauge_ax.set_xlim(start, end)
    formatter = mb_float_formatter if float_formatter else mb_formatter
    gauge_ax.xaxis.set_major_formatter(formatter)
    gauge_ax.xaxis.set_ticks(np.arange(start + gauge_step, end, gauge_step))
    gauge_ax.yaxis.set_ticks([])


def main():
    p = OptionParser(__doc__)
    p.add_option("--order", help="The order to plot the tracks, comma-separated")
    opts, args, iopts = p.set_image_options()

    if len(args) != 3:
        sys.exit(not p.print_help())

    chr, sizes, datadir = args
    order = opts.order
    hlsuffix = opts.hlsuffix
    if order:
        order = order.split(",")
    sizes = Sizes(sizes)
    fig = plt.figure(1, (iopts.w, iopts.h))
    root = fig.add_axes([0, 0, 1, 1])
    canvas = (0.12, 0.35, 0.8, 0.35)
    chr_size = sizes.get_size(chr)
    Coverage(
        fig, root, canvas, chr, (0, chr_size), datadir, order=order, hlsuffix=hlsuffix
    )

    root.set_xlim(0, 1)
    root.set_ylim(0, 1)
    root.set_axis_off()

    image_name = chr + "." + iopts.format
    savefig(image_name, dpi=iopts.dpi, iopts=iopts)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Run quality control (QC) on gene annotation. MAKER output was used during
testing. Several aspects of annotation QC are implemented in this script.

- Trim UTRs. MAKER sometimes predict UTRs that extend into other genes.
- Remove overlapping models.
"""
import sys

from jcvi.formats.gff import (
    Gff,
    get_piles,
    make_index,
    import_feats,
    populate_children,
    to_range,
)
from jcvi.formats.base import must_open
from jcvi.formats.sizes import Sizes
from jcvi.utils.range import range_minmax, range_chain, range_overlap
from jcvi.apps.base import OptionParser, ActionDispatcher


def main():

    actions = (
        ("trimUTR", "remove UTRs in the annotation set"),
        ("uniq", "remove overlapping gene models"),
        ("nmd", "identify transcript variant candidates for nonsense-mediated decay"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def uniq(args):
    """
    %prog uniq gffile cdsfasta

    Remove overlapping gene models. Similar to formats.gff.uniq(), overlapping
    'piles' are processed, one by one.

    Here, we use a different algorithm, that retains the best non-overlapping
    subset witin each pile, rather than single best model. Scoring function is
    also different, rather than based on score or span, we optimize for the
    subset that show the best combined score. Score is defined by:

    score = (1 - AED) * length
    """

    p = OptionParser(uniq.__doc__)
    p.set_outfile()
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    gffile, cdsfasta = args
    gff = Gff(gffile)
    sizes = Sizes(cdsfasta).mapping
    gene_register = {}
    for g in gff:
        if g.type != "mRNA":
            continue
        aed = float(g.attributes["_AED"][0])
        gene_register[g.parent] = (1 - aed) * sizes[g.accn]

    allgenes = import_feats(gffile)
    g = get_piles(allgenes)

    bestids = set()
    for group in g:
        ranges = [to_range(x, score=gene_register[x.accn], id=x.accn) for x in group]
        selected_chain, score = range_chain(ranges)
        bestids |= set(x.id for x in selected_chain)

    removed = set(x.accn for x in allgenes) - bestids
    fw = open("removed.ids", "w")
    print("\n".join(sorted(removed)), file=fw)
    fw.close()
    populate_children(opts.outfile, bestids, gffile, "gene")


def get_cds_minmax(g, cid, level=2):
    cds = [x for x in g.children(cid, level) if x.featuretype == "CDS"]
    cdsranges = [(x.start, x.end) for x in cds]
    return range_minmax(cdsranges)


def trim(c, start, end, trim5=False, trim3=False, both=True):
    cstart, cend = c.start, c.end
    # Trim coordinates for feature c based on overlap to start and end
    if ((trim5 or both) and c.strand == "+") or ((trim3 or both) and c.strand == "-"):
        c.start = max(cstart, start)
    if ((trim3 or both) and c.strand == "+") or ((trim5 or both) and c.strand == "-"):
        c.end = min(cend, end)

    if c.start != cstart or c.end != cend:
        print(
            c.id,
            "trimmed [{0}, {1}] => [{2}, {3}]".format(cstart, cend, c.start, c.end),
            file=sys.stderr,
        )
    else:
        print(c.id, "no change", file=sys.stderr)


def reinstate(c, rc, trim5=False, trim3=False, both=True):
    cstart, cend = c.start, c.end
    # reinstate coordinates for feature `c` based on reference feature `refc`
    if ((trim5 or both) and c.strand == "+") or ((trim3 or both) and c.strand == "-"):
        c.start = rc.start
    if ((trim3 or both) and c.strand == "+") or ((trim5 or both) and c.strand == "-"):
        c.end = rc.end

    if c.start != cstart or c.end != cend:
        print(
            c.id,
            "reinstated [{0}, {1}] => [{2}, {3}]".format(cstart, cend, c.start, c.end),
            file=sys.stderr,
        )
    else:
        print(c.id, "no change", file=sys.stderr)


def cmp_children(cid, gff, refgff, cftype="CDS"):
    start, end = get_cds_minmax(gff, cid, level=1)
    rstart, rend = get_cds_minmax(refgff, cid, level=1)
    return (
        ((start == rstart) and (end == rend))
        and (
            len(list(gff.children(cid, featuretype=cftype)))
            == len(list(refgff.children(cid, featuretype=cftype)))
        )
        and (
            gff.children_bp(cid, child_featuretype=cftype)
            == refgff.children_bp(cid, child_featuretype=cftype)
        )
    )


def fprint(c, fw):
    if c.start > c.end:
        print(c.id, "destroyed [{0} > {1}]".format(c.start, c.end), file=sys.stderr)
    else:
        print(c, file=fw)


def trimUTR(args):
    """
    %prog trimUTR gffile

    Remove UTRs in the annotation set.

    If reference GFF3 is provided, reinstate UTRs from reference
    transcripts after trimming.

    Note: After running trimUTR, it is advised to also run
    `python -m jcvi.formats.gff fixboundaries` on the resultant GFF3
    to adjust the boundaries of all parent 'gene' features
    """
    import gffutils
    from jcvi.formats.base import SetFile

    p = OptionParser(trimUTR.__doc__)
    p.add_option(
        "--trim5",
        default=None,
        type="str",
        help="File containing gene list for 5' UTR trimming",
    )
    p.add_option(
        "--trim3",
        default=None,
        type="str",
        help="File containing gene list for 3' UTR trimming",
    )
    p.add_option(
        "--trimrange",
        default=None,
        type="str",
        help="File containing gene list for UTR trim back"
        + "based on suggested (start, stop) coordinate range",
    )
    p.add_option(
        "--refgff",
        default=None,
        type="str",
        help="Reference GFF3 used as fallback to replace UTRs",
    )
    p.set_outfile()
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (gffile,) = args
    gff = make_index(gffile)

    trim_both = False if (opts.trim5 or opts.trim3) else True
    trim5 = SetFile(opts.trim5) if opts.trim5 else set()
    trim3 = SetFile(opts.trim3) if opts.trim3 else set()
    trimrange = dict()
    if opts.trimrange:
        trf = must_open(opts.trimrange)
        for tr in trf:
            assert (
                len(tr.split("\t")) == 3
            ), "Must specify (start, stop) coordinate range"
            id, start, stop = tr.split("\t")
            trimrange[id] = (int(start), int(stop))
        trf.close()

    refgff = make_index(opts.refgff) if opts.refgff else None

    fw = must_open(opts.outfile, "w")
    for feat in gff.iter_by_parent_childs(
        featuretype="gene", order_by=("seqid", "start"), level=1
    ):
        for c in feat:
            cid, ctype, cparent = (
                c.id,
                c.featuretype,
                c.attributes.get("Parent", [None])[0],
            )
            t5, t3 = False, False
            if ctype == "gene":
                t5 = True if cid in trim5 else False
                t3 = True if cid in trim3 else False
                start, end = get_cds_minmax(gff, cid)
                trim(c, start, end, trim5=t5, trim3=t3, both=trim_both)
                fprint(c, fw)
            elif ctype == "mRNA":
                utr_types, extras = [], set()
                if any(id in trim5 for id in (cid, cparent)):
                    t5 = True
                    trim5.add(cid)
                if any(id in trim3 for id in (cid, cparent)):
                    t3 = True
                    trim3.add(cid)
                refc = None
                if refgff:
                    try:
                        refc = refgff[cid]
                        refctype = refc.featuretype
                        refptype = refgff[refc.attributes["Parent"][0]].featuretype
                        if refctype == "mRNA" and refptype == "gene":
                            if cmp_children(cid, gff, refgff, cftype="CDS"):
                                reinstate(c, refc, trim5=t5, trim3=t3, both=trim_both)
                                if t5:
                                    utr_types.append("five_prime_UTR")
                                if t3:
                                    utr_types.append("three_prime_UTR")
                                for utr_type in utr_types:
                                    for utr in refgff.children(
                                        refc, featuretype=utr_type
                                    ):
                                        extras.add(utr)
                                        for exon in refgff.region(
                                            region=utr, featuretype="exon"
                                        ):
                                            if exon.attributes["Parent"][0] == cid:
                                                extras.add(exon)
                        else:
                            refc = None
                    except gffutils.exceptions.FeatureNotFoundError:
                        pass
                start, end = get_cds_minmax(gff, cid, level=1)
                if cid in trimrange:
                    start, end = range_minmax([trimrange[cid], (start, end)])
                if not refc:
                    trim(c, start, end, trim5=t5, trim3=t3, both=trim_both)
                fprint(c, fw)
                for cc in gff.children(cid, order_by="start"):
                    _ctype = cc.featuretype
                    if _ctype not in utr_types:
                        if _ctype != "CDS":
                            if _ctype == "exon":
                                eskip = [
                                    range_overlap(to_range(cc), to_range(x))
                                    for x in extras
                                    if x.featuretype == "exon"
                                ]
                                if any(eskip):
                                    continue
                            trim(cc, start, end, trim5=t5, trim3=t3, both=trim_both)
                            fprint(cc, fw)
                        else:
                            fprint(cc, fw)
                for x in extras:
                    fprint(x, fw)
    fw.close()


def nmd(args):
    """
    %prog nmd gffile

    Identify transcript variants which might be candidates for nonsense
    mediated decay (NMD)

    A transcript is considered to be a candidate for NMD when the CDS stop
    codon is located more than 50nt upstream of terminal splice site donor

    References:
    http://www.nature.com/horizon/rna/highlights/figures/s2_spec1_f3.html
    http://www.biomedcentral.com/1741-7007/7/23/figure/F1
    """
    from jcvi.utils.cbook import enumerate_reversed

    p = OptionParser(nmd.__doc__)
    p.set_outfile()

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (gffile,) = args
    gff = make_index(gffile)

    fw = must_open(opts.outfile, "w")
    for gene in gff.features_of_type("gene", order_by=("seqid", "start")):
        _enumerate = enumerate if gene.strand == "-" else enumerate_reversed
        for mrna in gff.children(gene, featuretype="mRNA", order_by="start"):
            tracker = dict()
            tracker["exon"] = list(
                gff.children(mrna, featuretype="exon", order_by="start")
            )
            tracker["cds"] = [None] * len(tracker["exon"])

            tcds_pos = None
            for i, exon in _enumerate(tracker["exon"]):
                for cds in gff.region(
                    region=exon, featuretype="CDS", completely_within=True
                ):
                    if mrna.id in cds["Parent"]:
                        tracker["cds"][i] = cds
                        tcds_pos = i
                        break
                if tcds_pos:
                    break

            NMD, distance = False, 0
            if (mrna.strand == "+" and tcds_pos + 1 < len(tracker["exon"])) or (
                mrna.strand == "-" and tcds_pos - 1 >= 0
            ):
                tcds = tracker["cds"][tcds_pos]
                texon = tracker["exon"][tcds_pos]

                PTC = tcds.end if mrna.strand == "+" else tcds.start
                TDSS = texon.end if mrna.strand == "+" else texon.start
                distance = abs(TDSS - PTC)
                NMD = True if distance > 50 else False

            print(
                "\t".join(
                    str(x)
                    for x in (
                        gene.id,
                        mrna.id,
                        gff.children_bp(mrna, child_featuretype="CDS"),
                        distance,
                        NMD,
                    )
                ),
                file=fw,
            )

    fw.close()


if __name__ == "__main__":
    main()



"""
Wrapper for fetching data from various online repositories \
(Entrez, Ensembl, Phytozome, and SRA)
"""
import logging
import os.path as op
import re
import sys
import time

from os.path import join as urljoin
from urllib.error import HTTPError, URLError

from Bio import Entrez, SeqIO
from more_itertools import grouper

from jcvi.formats.base import FileShredder, must_open
from jcvi.formats.fasta import print_first_difference
from jcvi.formats.fastq import fromsra
from jcvi.utils.cbook import tile
from jcvi.utils.console import printf
from jcvi.apps.base import (
    OptionParser,
    ActionDispatcher,
    get_email_address,
    mkdir,
    ls_ftp,
    download,
    sh,
    last_updated,
    which,
)


myEmail = get_email_address()
Entrez.email = myEmail
PHYTOZOME_COOKIES = ".phytozome_cookies"


def batch_taxonomy(list_of_taxids):
    """
    Convert list of taxids to Latin names
    """
    for taxid in list_of_taxids:
        handle = Entrez.efetch(db="Taxonomy", id=taxid, retmode="xml")
        records = Entrez.read(handle)
        yield records[0]["ScientificName"]


def batch_taxids(list_of_names):
    """
    Opposite of batch_taxonomy():

    Convert list of Latin names to taxids
    """
    for name in list_of_names:
        handle = Entrez.esearch(db="Taxonomy", term=name, retmode="xml")
        records = Entrez.read(handle)
        yield records["IdList"][0]


def batch_entrez(
    list_of_terms, db="nuccore", retmax=1, rettype="fasta", batchsize=1, email=myEmail
):
    """
    Retrieve multiple rather than a single record
    """

    for term in list_of_terms:

        logging.debug("Search term %s", term)
        success = False
        ids = None
        if not term:
            continue

        while not success:
            try:
                search_handle = Entrez.esearch(db=db, retmax=retmax, term=term)
                rec = Entrez.read(search_handle)
                success = True
                ids = rec["IdList"]
            except (HTTPError, URLError, RuntimeError, KeyError) as e:
                logging.error(e)
                logging.debug("wait 5 seconds to reconnect...")
                time.sleep(5)

        if not ids:
            logging.error("term {0} not found".format(term))
            continue

        assert ids
        nids = len(ids)
        if nids > 1:
            logging.debug("A total of {0} results found.".format(nids))

        if batchsize != 1:
            logging.debug("Use a batch size of {0}.".format(batchsize))

        ids = list(grouper(ids, batchsize))

        for id in ids:
            id = [x for x in id if x]
            size = len(id)
            id = ",".join(id)

            success = False
            while not success:
                try:
                    fetch_handle = Entrez.efetch(
                        db=db, id=id, rettype=rettype, email=email
                    )
                    success = True
                except (HTTPError, URLError, RuntimeError) as e:
                    logging.error(e)
                    logging.debug("wait 5 seconds to reconnect...")
                    time.sleep(5)

            yield id, size, term, fetch_handle


def main():

    actions = (
        ("entrez", "fetch records from entrez using a list of GenBank accessions"),
        ("bisect", "determine the version of the accession by querying entrez"),
        (
            "phytozome9",
            "retrieve genomes and annotations from phytozome version 9.0 (legacy)",
        ),
        ("phytozome", "retrieve genomes and annotations from phytozome"),
        ("ensembl", "retrieve genomes and annotations from ensembl"),
        ("sra", "retrieve files from SRA via the sra-instant FTP"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def ensembl(args):
    """
    %prog ensembl species

    Retrieve genomes and annotations from ensembl FTP. Available species
    listed below. Use comma to give a list of species to download. For example:

    $ %prog ensembl danio_rerio,gasterosteus_aculeatus
    """
    p = OptionParser(ensembl.__doc__)
    p.add_option("--version", default="75", help="Ensembl version")
    opts, args = p.parse_args(args)

    version = opts.version
    url = "ftp://ftp.ensembl.org/pub/release-{0}/".format(version)
    fasta_url = url + "fasta/"

    valid_species = [x for x in ls_ftp(fasta_url) if "." not in x]
    doc = "\n".join((ensembl.__doc__, tile(valid_species)))
    p.set_usage(doc)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (species,) = args
    species = species.split(",")
    for s in species:
        download_species_ensembl(s, valid_species, url)


def download_species_ensembl(species, valid_species, url):
    assert species in valid_species, "{0} is not in the species list".format(species)

    # We want to download assembly and annotation for given species
    ann_url = urljoin(url, "gtf/{0}".format(species))
    cds_url = urljoin(url, "fasta/{0}/cds".format(species))

    for u in (ann_url, cds_url):
        valid_files = [x for x in ls_ftp(u) if x.endswith(".gz")]
        for f in valid_files:
            f = urljoin(u, f)
            download(f)


def get_cookies(cookies=PHYTOZOME_COOKIES):
    from jcvi.utils.console import console

    # Check if cookies is still good
    if op.exists(cookies) and last_updated(cookies) < 3600:
        return cookies

    if console.is_terminal:
        username = console.input("[bold green]Phytozome Login: ")
        pw = console.input("[bold green]Phytozome Password: ", password=True)
    else:
        username, pw = None, None
    curlcmd = which("curl")
    if curlcmd is None:
        logging.error("curl command not installed. Aborting.")
        return None
    cmd = "{} https://signon.jgi.doe.gov/signon/create".format(curlcmd)
    cmd += " --data-urlencode 'login={0}' --data-urlencode 'password={1}' -b {2} -c {2}".format(
        username, pw, cookies
    )
    sh(cmd, outfile="/dev/null", errfile="/dev/null", log=False)
    if not op.exists(cookies):
        logging.error("Cookies file `{}` not created. Aborting.".format(cookies))
        return None

    return cookies


def phytozome(args):
    """
    %prog phytozome species

    Retrieve genomes and annotations from phytozome using Globus API. Available
    species listed below. Use comma to give a list of species to download. For
    example:

    $ %prog phytozome Athaliana,Vvinifera,Osativa,Sbicolor,Slycopersicum

    The downloader will prompt you to enter Phytozome user name and password
    during downloading. Please register for a login at:
    https://phytozome.jgi.doe.gov/pz/portal.html.
    """
    from jcvi.apps.biomart import GlobusXMLParser

    p = OptionParser(phytozome.__doc__)
    p.add_option(
        "--version",
        default="12",
        choices=("9", "10", "11", "12", "12_unrestricted", "13"),
        help="Phytozome version",
    )
    p.add_option(
        "--assembly",
        default=False,
        action="store_true",
        help="Download assembly",
    )
    p.add_option(
        "--format",
        default=False,
        action="store_true",
        help="Format to CDS and BED for synteny inference",
    )
    p.set_downloader()
    opts, args = p.parse_args(args)

    downloader = opts.downloader
    directory_listing = ".phytozome_directory_V{}.xml".format(opts.version)
    # Get directory listing
    base_url = "http://genome.jgi.doe.gov"
    dlist = "{}/ext-api/downloads/get-directory?organism=PhytozomeV{}".format(
        base_url, opts.version
    )

    # Make sure we have a valid cookies
    cookies = get_cookies()
    if cookies is None:
        logging.error("Error fetching cookies ... cleaning up")
        FileShredder([directory_listing])
        sys.exit(1)

    # Proceed to use the cookies and download the species list
    try:
        download(
            dlist,
            filename=directory_listing,
            cookies=cookies,
            downloader=downloader,
        )
        g = GlobusXMLParser(directory_listing)
    except:
        logging.error("Error downloading directory listing ... cleaning up")
        FileShredder([directory_listing, cookies])
        sys.exit(1)

    genomes = g.get_genomes()
    valid_species = genomes.keys()
    species_tile = tile(valid_species)
    p.set_usage("\n".join((phytozome.__doc__, species_tile)))

    if len(args) != 1:
        sys.exit(not p.print_help())

    (species,) = args
    if species == "all":
        species = ",".join(valid_species)

    species = species.split(",")
    for s in species:
        res = download_species_phytozome(
            genomes,
            s,
            valid_species,
            base_url,
            cookies,
            assembly=opts.assembly,
            downloader=downloader,
        )
        if not res:
            logging.error("No files downloaded")
        gff, fa = res.get("gff"), res.get("cds")
        if opts.format:
            format_bed_and_cds(s, gff, fa)


def download_species_phytozome(
    genomes, species, valid_species, base_url, cookies, assembly=False, downloader=None
):
    """Download assembly FASTA and annotation GFF.

    Args:
        genomes (dict): Dictionary parsed from Globus XML.
        species (str): Target species to download.
        valid_species (List[str]): Allowed set of species
        base_url (str): URL.
        cookies (str): cookies file path.
        assembly (bool, optional): Do we download assembly FASTA (can be big).
        Defaults to False.
        downloader (str, optional): Use a given downloader. One of wget|curl|powershell|insecure.
        Defaults to None.
    """
    assert species in valid_species, "{} is not in the species list".format(species)
    res = {}
    genome = genomes.get(species)
    if not genome:
        return res

    genome_assembly = genome.get("assembly")
    if assembly and genome_assembly:
        asm_name = next(x for x in genome_assembly if x.endswith(".fa.gz"))
        if asm_name:
            res["asm"] = genome_assembly.download(
                asm_name, base_url, cookies, downloader=downloader
            )

    genome_annotation = genome.get("annotation")
    if genome_annotation:
        gff_name = next(x for x in genome_annotation if x.endswith(".gene.gff3.gz"))
        if gff_name:
            res["gff"] = genome_annotation.download(
                gff_name, base_url, cookies, downloader=downloader
            )
        cds_name = next(x for x in genome_annotation if x.endswith(".cds.fa.gz"))
        if cds_name:
            res["cds"] = genome_annotation.download(
                cds_name, base_url, cookies, downloader=downloader
            )

    return res


def phytozome9(args):
    """
    %prog phytozome9 species

    Retrieve genomes and annotations from phytozome FTP. Available species
    listed below. Use comma to give a list of species to download. For example:

    $ %prog phytozome9 Athaliana,Vvinifera,Osativa,Sbicolor,Slycopersicum
    """
    p = OptionParser(phytozome9.__doc__)
    p.add_option(
        "--assembly",
        default=False,
        action="store_true",
        help="Download assembly",
    )
    p.add_option(
        "--format",
        default=False,
        action="store_true",
        help="Format to CDS and BED for synteny inference",
    )
    opts, args = p.parse_args(args)

    version = "9.0"
    url = "ftp://ftp.jgi-psf.org/pub/compgen/phytozome/v{0}/".format(version)
    valid_species = [x for x in ls_ftp(url) if "." not in x]

    doc = "\n".join((phytozome9.__doc__, tile(valid_species)))
    p.set_usage(doc)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (species,) = args
    if species == "all":
        species = ",".join(valid_species)

    species = species.split(",")

    for s in species:
        res = download_species_phytozome9(s, valid_species, url, assembly=opts.assembly)
        if not res:
            logging.error("No files downloaded")
        gff, cdsfa = res.get("gff"), res.get("cds")
        if opts.format:
            format_bed_and_cds(s, gff, cdsfa)


def format_bed_and_cds(species, gff, cdsfa):
    """Run gff.format() and fasta.format() to generate BED and CDS files.
    This prepares the input files for the MCscan synteny workflow.

    https://github.com/tanghaibao/jcvi/wiki/MCscan-(Python-version)

    Args:
        species (str): Name of the species
        gff (str): Path to the GFF file
        cdsfa (str): Path to the FASTA file
    """
    from jcvi.formats.gff import bed as gff_bed
    from jcvi.formats.fasta import format as fasta_format

    # We have to watch out when the gene names and mRNA names mismatch, in which
    # case we just extract the mRNA names
    use_IDs = set()
    use_mRNAs = {
        "Cclementina",
        "Creinhardtii",
        "Csinensis",
        "Fvesca",
        "Lusitatissimum",
        "Mesculenta",
        "Mguttatus",
        "Ppersica",
        "Pvirgatum",
        "Rcommunis",
        "Sitalica",
        "Tcacao",
        "Thalophila",
        "Vcarteri",
        "Vvinifera",
        "Zmays",
    }
    key = "ID" if species in use_IDs else "Name"
    ttype = "mRNA" if species in use_mRNAs else "gene"
    bedfile = species + ".bed"
    cdsfile = species + ".cds"
    gff_bed([gff, "--type={}".format(ttype), "--key={}".format(key), "-o", bedfile])
    fasta_format([cdsfa, cdsfile, r"--sep=|"])


def download_species_phytozome9(species, valid_species, base_url, assembly=False):
    assert species in valid_species, "{} is not in the species list".format(species)

    # We want to download assembly and annotation for given species
    surl = urljoin(base_url, species)
    contents = [x for x in ls_ftp(surl) if x.endswith("_readme.txt")]
    magic = contents[0].split("_")[1]  # Get the magic number
    logging.debug("Found magic number for {0}: {1}".format(species, magic))

    pf = "{0}_{1}".format(species, magic)
    asm_url = urljoin(surl, "assembly/{0}.fa.gz".format(pf))
    ann_url = urljoin(surl, "annotation/{0}_gene.gff3.gz".format(pf))
    cds_url = urljoin(surl, "annotation/{0}_cds.fa.gz".format(pf))
    res = {}
    if assembly:
        res["asm"] = download(asm_url)
    res["gff"] = download(ann_url)
    res["cds"] = download(cds_url)
    return res


def get_first_rec(fastafile):
    """
    Returns the first record in the fastafile
    """
    f = list(SeqIO.parse(fastafile, "fasta"))

    if len(f) > 1:
        logging.debug(
            "{0} records found in {1}, using the first one".format(len(f), fastafile)
        )

    return f[0]


def bisect(args):
    """
    %prog bisect acc accession.fasta

    determine the version of the accession by querying entrez, based on a fasta file.
    This proceeds by a sequential search from xxxx.1 to the latest record.
    """
    p = OptionParser(bisect.__doc__)
    p.set_email()

    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    acc, fastafile = args
    arec = get_first_rec(fastafile)

    valid = None
    for i in range(1, 100):
        term = "%s.%d" % (acc, i)
        try:
            query = list(batch_entrez([term], email=opts.email))
        except AssertionError as e:
            logging.debug(f"no records found for {term}. terminating. {e}")
            return

        id, term, handle = query[0]
        brec = next(SeqIO.parse(handle, "fasta"))

        match = print_first_difference(
            arec, brec, ignore_case=True, ignore_N=True, rc=True
        )
        if match:
            valid = term
            break

    if valid:
        printf()
        printf("[green]{} matches the sequence in `{}`".format(valid, fastafile))


def entrez(args):
    """
    %prog entrez <filename|term>

    `filename` contains a list of terms to search. Or just one term. If the
    results are small in size, e.g. "--format=acc", use "--batchsize=100" to speed
    the download.
    """
    p = OptionParser(entrez.__doc__)

    allowed_databases = {
        "fasta": ["genome", "nuccore", "nucgss", "protein", "nucest"],
        "asn.1": ["genome", "nuccore", "nucgss", "protein", "gene"],
        "xml": ["genome", "nuccore", "nucgss", "nucest", "gene"],
        "gb": ["genome", "nuccore", "nucgss"],
        "est": ["nucest"],
        "gss": ["nucgss"],
        "acc": ["nuccore"],
    }

    valid_formats = tuple(allowed_databases.keys())
    valid_databases = ("genome", "nuccore", "nucest", "nucgss", "protein", "gene")

    p.add_option(
        "--noversion",
        dest="noversion",
        default=False,
        action="store_true",
        help="Remove trailing accession versions",
    )
    p.add_option(
        "--format",
        default="fasta",
        choices=valid_formats,
        help="download format",
    )
    p.add_option(
        "--database",
        default="nuccore",
        choices=valid_databases,
        help="search database",
    )
    p.add_option(
        "--retmax",
        default=1000000,
        type="int",
        help="how many results to return",
    )
    p.add_option(
        "--skipcheck",
        default=False,
        action="store_true",
        help="turn off prompt to check file existence",
    )
    p.add_option(
        "--batchsize",
        default=500,
        type="int",
        help="download the results in batch for speed-up",
    )
    p.set_outdir(outdir=None)
    p.add_option("--outprefix", default="out", help="output file name prefix")
    p.set_email()
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(p.print_help())

    (filename,) = args
    if op.exists(filename):
        pf = filename.rsplit(".", 1)[0]
        list_of_terms = [row.strip() for row in open(filename)]
        if opts.noversion:
            list_of_terms = [x.rsplit(".", 1)[0] for x in list_of_terms]
    else:
        pf = filename
        # the filename is the search term
        list_of_terms = [filename.strip()]

    fmt = opts.format
    database = opts.database
    batchsize = opts.batchsize

    assert (
        database in allowed_databases[fmt]
    ), "For output format '{0}', allowed databases are: {1}".format(
        fmt, allowed_databases[fmt]
    )
    assert batchsize >= 1, "batchsize must >= 1"

    if " " in pf:
        pf = opts.outprefix

    outfile = "{0}.{1}".format(pf, fmt)

    outdir = opts.outdir
    if outdir:
        mkdir(outdir)

    # If noprompt, will not check file existence
    if not outdir:
        fw = must_open(outfile, "w", checkexists=True, skipcheck=opts.skipcheck)
        if fw is None:
            return

    seen = set()
    totalsize = 0
    for id, size, term, handle in batch_entrez(
        list_of_terms,
        retmax=opts.retmax,
        rettype=fmt,
        db=database,
        batchsize=batchsize,
        email=opts.email,
    ):
        if outdir:
            outfile = urljoin(outdir, "{0}.{1}".format(term, fmt))
            fw = must_open(outfile, "w", checkexists=True, skipcheck=opts.skipcheck)
            if fw is None:
                continue

        rec = handle.read()
        if id in seen:
            logging.error("Duplicate key ({0}) found".format(rec))
            continue

        totalsize += size
        print(rec, file=fw)
        print(file=fw)

        seen.add(id)

    if seen:
        printf(
            "A total of {0} {1} records downloaded.".format(totalsize, fmt.upper()),
        )

    return outfile


def sra(args):
    """
    %prog sra [term|term.ids]

    Given an SRA run ID, fetch the corresponding .sra file from the sra-instant FTP.
    The term can also be a file containing list of SRR ids, one per line.

    Once downloaded, the SRA file is processed through `fastq-dump` to produce
    FASTQ formatted sequence files, which are gzipped by default.
    """
    p = OptionParser(sra.__doc__)

    p.add_option(
        "--nogzip",
        dest="nogzip",
        default=False,
        action="store_true",
        help="Do not gzip the FASTQ generated by fastq-dump",
    )
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (term,) = args
    if op.isfile(term):
        terms = [x.strip() for x in open(term)]
    else:
        terms = [term]

    for term in terms:
        srafile = download_srr_term(term)
        pf = srafile.split(".")[0]
        mkdir(pf)
        _opts = [srafile, "--paired", "--outdir={0}".format(pf)]
        if not opts.nogzip:
            _opts.append("--compress=gzip")
        fromsra(_opts)


def download_srr_term(term):
    sra_base_url = "ftp://ftp-trace.ncbi.nlm.nih.gov/sra/sra-instant/reads/ByRun/sra/"
    sra_run_id_re = re.compile(r"^([DES]RR)(\d{3})(\d{3,4})$")

    m = re.search(sra_run_id_re, term)
    if m is None:
        logging.error(
            "Incorrect SRA identifier format "
            + "[should be like SRR126150, SRR1001901. "
            + "len(identifier) should be between 9-10 characters]"
        )
        sys.exit()

    prefix, subprefix = m.group(1), "{0}{1}".format(m.group(1), m.group(2))
    download_url = urljoin(
        sra_base_url, prefix, subprefix, term, "{0}.sra".format(term)
    )

    logging.debug("Downloading file: {0}".format(download_url))
    return download(download_url)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
TSP solver using Concorde or OR-tools. This is much faster than the LP-formulation in
algorithms.lpsolve.tsp(). See also:
https://developers.google.com/optimization/routing/tsp
"""
import os.path as op
import os
import logging
import shutil
import numpy as np

from collections import defaultdict
from dataclasses import dataclass
from itertools import combinations
from more_itertools import pairwise

from jcvi.formats.base import FileShredder, must_open
from jcvi.apps.base import mkdir, which, sh


INF = 10000
NEG_INF = -INF
Work_dir = "tsp_work"


@dataclass
class TSPDataModel:
    edges: list  # List of tuple (source, target, weight)

    def distance_matrix(self, precision=0) -> tuple:
        """Compute the distance matrix

        Returns:
            np.array: Numpy square matrix with integer entries as distance
        """
        _, _, nodes = node_to_edge(self.edges, directed=False)
        nodes_indices = dict((n, i) for i, n in enumerate(nodes))
        nnodes = len(nodes)

        # TSPLIB requires explicit weights to be integral, and non-negative
        weights = [x[-1] for x in self.edges]
        max_x, min_x = max(weights), min(weights)
        inf = 2 * max(abs(max_x), abs(min_x))
        factor = 10 ** precision
        logging.debug(
            "TSP rescale: max_x=%d, min_x=%d, inf=%d, factor=%d",
            max_x,
            min_x,
            inf,
            factor,
        )

        D = np.ones((nnodes, nnodes), dtype=float) * inf
        for a, b, w in self.edges:
            ia, ib = nodes_indices[a], nodes_indices[b]
            D[ia, ib] = D[ib, ia] = w
        D = (D - min_x) * factor
        D = D.astype(int)
        return D, nodes

    def solve(self, time_limit=5, concorde=False, precision=0) -> list:
        """Solve the TSP instance.

        Args:
            time_limit (int, optional): Time limit to run. Default to 5 seconds.
            concorde (bool, optional): Shall we run concorde? Defaults to False.
            precision (int, optional): Float precision of distance. Defaults to 0.

        Returns:
            list: Ordered list of node indices to visit
        """
        if concorde:
            return Concorde(self, precision=precision).tour

        # Use OR-tools
        from ortools.constraint_solver import routing_enums_pb2
        from ortools.constraint_solver import pywrapcp

        D, nodes = self.distance_matrix(precision)
        nnodes = len(nodes)

        # Create the routing index manager
        manager = pywrapcp.RoutingIndexManager(nnodes, 1, 0)

        # Create routing model
        routing = pywrapcp.RoutingModel(manager)

        def distance_callback(from_index, to_index):
            """Returns the distance between the two nodes."""
            from_node = manager.IndexToNode(from_index)
            to_node = manager.IndexToNode(to_index)
            return D[from_node, to_node]

        transit_callback_index = routing.RegisterTransitCallback(distance_callback)

        # Define cost of each arc
        routing.SetArcCostEvaluatorOfAllVehicles(transit_callback_index)

        # Search strategy
        search_parameters = pywrapcp.DefaultRoutingSearchParameters()
        search_parameters.local_search_metaheuristic = (
            routing_enums_pb2.LocalSearchMetaheuristic.GUIDED_LOCAL_SEARCH
        )
        search_parameters.time_limit.seconds = time_limit

        # Solve the problem
        solution = routing.SolveWithParameters(search_parameters)

        tour = []
        logging.info("Objective: %d", solution.ObjectiveValue())
        index = routing.Start(0)
        route_distance = 0
        while not routing.IsEnd(index):
            tour.append(manager.IndexToNode(index))
            previous_index = index
            index = solution.Value(routing.NextVar(index))
            route_distance = routing.GetArcCostForVehicle(previous_index, index, 0)
        logging.info("Route distance: %d", route_distance)

        return [nodes[x] for x in tour]


class Concorde(object):
    def __init__(
        self,
        data: TSPDataModel,
        work_dir=Work_dir,
        clean=True,
        verbose=False,
        precision=0,
        seed=666,
    ):
        """Run concorde on TSP instance

        Args:
            data (TSPDataModel): TSP instance with edge weights
            work_dir ([type], optional): Path to the work dir. Defaults to Work_dir.
            clean (bool, optional): Clean up intermediate results. Defaults to True.
            verbose (bool, optional): Show verbose messages. Defaults to False.
            precision (int, optional): Float precision of distance. Defaults to 0.
            seed (int, optional): Random seed. Defaults to 666.
        """
        self.data = data
        self.work_dir = work_dir
        self.clean = clean
        self.verbose = verbose

        mkdir(work_dir)
        tspfile = op.join(work_dir, "data.tsp")
        self.print_to_tsplib(tspfile, precision=precision)
        _, outfile = self.run_concorde(tspfile, seed=seed)
        self.tour = self.parse_output(outfile)

        if clean:
            shutil.rmtree(work_dir)
            residual_output = ["data.sol", "data.res", "Odata.res"]
            FileShredder(residual_output, verbose=False)

    def print_to_tsplib(self, tspfile, precision=0):
        """
        See TSPlib format:
        <https://www.iwr.uni-heidelberg.de/groups/comopt/software/TSPLIB95/>

        NAME: bayg29
        TYPE: TSP
        COMMENT: 29 Cities in Bavaria, geographical distances
        DIMENSION: 29
        EDGE_WEIGHT_TYPE: EXPLICIT
        EDGE_WEIGHT_FORMAT: UPPER_ROW
        DISPLAY_DATA_TYPE: TWOD_DISPLAY
        EDGE_WEIGHT_SECTION
        (... numbers ...)
        """
        fw = must_open(tspfile, "w")
        D, nodes = self.data.distance_matrix(precision)
        self.nodes = nodes
        self.nnodes = len(nodes)

        print("NAME: data", file=fw)
        print("TYPE: TSP", file=fw)
        print("DIMENSION: {}".format(self.nnodes), file=fw)
        print("EDGE_WEIGHT_TYPE: EXPLICIT", file=fw)
        print("EDGE_WEIGHT_FORMAT: FULL_MATRIX", file=fw)
        print("EDGE_WEIGHT_SECTION", file=fw)

        for row in D:  # Dump the full matrix
            print(" " + " ".join(str(x) for x in row), file=fw)

        print("EOF", file=fw)
        fw.close()
        logging.debug("Write TSP instance to `%s`", tspfile)

    def run_concorde(self, tspfile, seed=666):
        outfile = op.join(self.work_dir, "data.sol")
        if op.exists(outfile):
            os.remove(outfile)

        cc = "concorde"
        assert which(cc), (
            "You must install `concorde` on your PATH"
            + " [http://www.math.uwaterloo.ca/tsp/concorde.html]"
        )
        cmd = "{0} -s {1} -x -o {2} {3}".format(cc, seed, outfile, tspfile)

        outf = None if self.verbose else "/dev/null"
        retcode = sh(cmd, outfile=outf, errfile=outf)
        return retcode, outfile

    def parse_output(self, outfile):
        fp = open(outfile)
        dimension = int(next(fp).strip())  # header
        assert dimension == self.nnodes
        tour = []
        for row in fp:
            tour += [int(x) for x in row.split()]
        tour = [self.nodes[x] for x in tour]
        return tour


def node_to_edge(edges, directed=True):
    """
    From list of edges, record per node, incoming and outgoing edges
    """
    outgoing = defaultdict(set)
    incoming = defaultdict(set) if directed else outgoing
    nodes = set()
    for i, edge in enumerate(edges):
        (
            a,
            b,
        ) = edge[:2]
        outgoing[a].add(i)
        incoming[b].add(i)
        nodes.add(a)
        nodes.add(b)
    nodes = list(nodes)
    return outgoing, incoming, nodes


def populate_edge_weights(edges):
    # assume weight is 1 if not specified
    new_edges = []
    for e in edges:
        assert len(e) in (2, 3)
        if len(e) == 2:
            a, b = e
            w = 1
        else:
            a, b, w = e
        new_edges.append((a, b, w))
    return new_edges


def hamiltonian(edges, directed=False, time_limit=5, concorde=False, precision=0):
    """
    Calculates shortest path that traverses each node exactly once. Convert
    Hamiltonian path problem to TSP by adding one dummy point that has a distance
    of zero to all your other points. Solve the TSP and get rid of the dummy
    point - what remains is the Hamiltonian Path.

    >>> g = [(1,2), (2,3), (3,4), (4,2), (3,5)]
    >>> hamiltonian(g)
    [1, 2, 4, 3, 5]
    >>> hamiltonian([(1, 2), (2, 3)], directed=True)
    [1, 2, 3]
    """
    edges = populate_edge_weights(edges)
    _, _, nodes = node_to_edge(edges, directed=False)
    DUMMY = "DUMMY"
    dummy_edges = edges + [(DUMMY, x, 0) for x in nodes]
    if directed:
        dummy_edges += [(x, DUMMY, 0) for x in nodes]
        dummy_edges = reformulate_atsp_as_tsp(dummy_edges)

    tour = tsp(
        dummy_edges, time_limit=time_limit, concorde=concorde, precision=precision
    )

    dummy_index = tour.index(DUMMY)
    tour = tour[dummy_index:] + tour[:dummy_index]
    if directed:
        dummy_star_index = tour.index((DUMMY, "*"))
        assert dummy_star_index in (1, len(tour) - 1), tour
        if dummy_star_index == len(tour) - 1:  # need to flip
            tour = tour[1:] + tour[:1]
            tour = tour[::-1]
        path = tour[1:]
        path = [x for x in path if not isinstance(x, tuple)]
    else:
        path = tour[1:]

    return path


def tsp(edges, time_limit=5, concorde=False, precision=0) -> list:
    """Compute TSP solution

    Args:
        edges (list): List of tuple (source, target, weight)
        time_limit (int, optional): Time limit to run. Default to 5 seconds.
        concorde (bool, optional): Shall we run concorde? Defaults to False.
        precision (int, optional): Float precision of distance. Defaults to 0.

    Returns:
        list: List of nodes to visit
    """
    data = TSPDataModel(edges)
    return data.solve(time_limit=time_limit, concorde=concorde, precision=precision)


def reformulate_atsp_as_tsp(edges):
    """
    To reformulate the ATSP as a TSP, for each city a dummy city (e.g, for New
    York, a dummy city New York* is added. Between each city and its
    corresponding dummy city a negative or very small distance with value cheap
    is used. This makes sure that each cities always occurs in the solution
    together with its dummy city. The original distances are used between the
    cities and the dummy cities, where each city is responsible for the distance
    going to the city and the dummy city is responsible for the distance coming
    from the city. The distances between all cities and the distances between
    all dummy cities are set to infeasible.
    """
    _, _, nodes = node_to_edge(edges, directed=False)
    new_edges = []
    for a, b, w in edges:
        new_edges.append(((a, "*"), b, w))
    for n in nodes:
        new_edges.append((n, (n, "*"), NEG_INF))  # A negative weight
    return new_edges


def make_data(N, directed=False):
    x = np.random.randn(N)
    y = np.random.randn(N)
    xy = list(zip(x, y))
    M = np.zeros((N, N), dtype=float)
    for ia, ib in combinations(range(N), 2):
        ax, ay = xy[ia]
        bx, by = xy[ib]
        d = ((ax - bx) ** 2 + (ay - by) ** 2) ** 0.5
        M[ia, ib] = M[ib, ia] = d

    edges = []
    for ia, ib in combinations(range(N), 2):
        edges.append((ia, ib, M[ia, ib]))
        if directed:
            edges.append((ib, ia, M[ib, ia]))

    return x, y, M, edges


def evaluate(tour, M):
    score = 0
    for ia, ib in pairwise(tour):
        score += M[ia, ib]
    return score


def plot_data(x, y, tour, M):
    from jcvi.graphics.base import plt, savefig

    plt.plot(x, y, "ro")
    for ia, ib in pairwise(tour):
        plt.plot((x[ia], x[ib]), (y[ia], y[ib]), "r-")

    score = evaluate(tour, M)
    plt.title("Score={0:.2f}".format(score))

    savefig("demo.pdf")


def concorde_demo(POINTS=100):
    x, y, M, edges = make_data(POINTS)
    ctour = hamiltonian(edges, precision=3)
    plot_data(x, y, ctour, M)


def compare_lpsolve_to_concorde(POINTS=80, directed=False):
    from jcvi.algorithms.lpsolve import hamiltonian as lhamiltonian

    _, _, M, edges = make_data(POINTS, directed=directed)
    ltour = lhamiltonian(edges, directed=directed)
    print(ltour, evaluate(ltour, M))

    ctour = hamiltonian(edges, directed=directed, precision=3)
    print(ctour, evaluate(ctour, M))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Identify repeat numbers in STR repeats.
"""
import re
import os
import os.path as op
import json
import sys

try:
    import vcf
except ImportError:
    pass

import logging
import pyfasta
import numpy as np
import pandas as pd

from math import log, ceil
from collections import Counter, defaultdict
from multiprocessing import Pool

from jcvi.utils.cbook import percentage, uniqify
from jcvi.formats.base import timestamp
from jcvi.formats.bed import natsorted
from jcvi.apps.grid import MakeManager
from jcvi.formats.base import LineFile, must_open
from jcvi.utils.aws import push_to_s3, pull_from_s3, check_exists_s3, ls_s3
from jcvi.apps.base import (
    OptionParser,
    ActionDispatcher,
    mkdir,
    need_update,
    datafile,
    sh,
)


REF = "hg38"
REPO = datafile("TREDs.meta.csv")

READLEN = 150
MINSCORE = 36
YSEARCH_HAPLOTYPE = """
DYS393  DYS390 DYS19/DYS394  DYS19b        DYS391        DYS385a       DYS385b DYS426  DYS388  DYS439
DYS389I DYS392 DYS389B       DYS458        DYS459a/b     DYS459a/b     DYS455  DYS454  DYS447  DYS437
DYS448  DYS449 DYS464a/b/c/d DYS464a/b/c/d DYS464a/b/c/d DYS464a/b/c/d DYS464e DYS464f DYS464g DYS460
GATA-H4 YCAIIa YCAIIb        DYS456        DYS607        DYS576        DYS570  CDYa    CDYb    DYS442
DYS438  DYS531 DYS578        DYS395S1a/b   DYS395S1a/b   DYS590        DYS537  DYS641  DYS472  DYS406S1
DYS511  DYS425 DYS413a       DYS413b       DYS557        DYS594        DYS436  DYS490  DYS534  DYS450
DYS444  DYS481 DYS520        DYS446        DYS617        DYS568        DYS487  DYS572  DYS640  DYS492
DYS565  DYS461 DYS462        GATA-A10      DYS635        GAAT1B07      DYS441  DYS445  DYS452  DYS463
DYS434  DYS435 DYS485        DYS494        DYS495        DYS505        DYS522  DYS533  DYS549  DYS556
DYS575  DYS589 DYS636        DYS638        DYS643        DYS714        DYS716  DYS717  DYS726  DXYS156-Y
""".split()
YSEARCH_LL = """
L1  L2  L3  L4  L5  L6  L7  L8  L9  L10
L11 L12 L13 L14 L15 L16 L17 L18 L19 L20
L21 L22 L23 L24 L25 L26 L27 L28 L29 L30
L31 L32 L33 L34 L35 L36 L37 L38 L39 L40
L41 L54 L55 L56 L57 L58 L59 L60 L61 L62
L63 L42 L64 L65 L66 L67 L68 L69 L70 L71
L49 L72 L73 L51 L74 L75 L76 L77 L78 L79
L80 L43 L44 L45 L46 L47 L48 L50 L52 L53
L81 L82 L83 L84 L85 L86 L87 L88 L89 L90
L91 L92 L93 L94 L95 L96 L97 L98 L99 L100
""".split()
YHRD_YFILER = """
DYS456 DYS389I DYS390 DYS389B DYS458 DYS19/DYS394 DYS385
DYS393 DYS391 DYS439 DYS635 DYS392 GATA-H4 DYS437 DYS438 DYS448
""".split()
YHRD_YFILERPLUS = """
DYS576 DYS389I DYS635 DYS389B DYS627 DYS460 DYS458 DYS19/DYS394 GATA-H4 DYS448 DYS391
DYS456 DYS390 DYS438 DYS392 DYS518 DYS570 DYS437 DYS385a DYS449
DYS393 DYS439 DYS481 DYF387S1 DYS533
""".split()
USYSTR_ALL = """
DYF387S1 DYS19/DYS394 DYS385 DYS389I
DYS389B DYS390 DYS391 DYS392
DYS393 DYS437 DYS438 DYS439
DYS448 DYS449 DYS456 DYS458
DYS460 DYS481 DYS518 DYS533
DYS549 DYS570 DYS576 DYS627
DYS635 DYS643 GATA-H4
""".split()


class TREDsRepo(dict):
    def __init__(self, ref=REF):
        super(TREDsRepo, self).__init__()
        self.ref = ref
        df = pd.read_csv(REPO, index_col=0)
        self.names = []
        for name, row in df.iterrows():
            self[name] = TRED(name, row, ref=ref)
            self.names.append(name)
        self.df = df

    def to_json(self):
        s = self.df.to_json(orient="index")
        s = s.decode("windows-1252").encode("utf8")
        s = json.dumps(json.loads(s), sort_keys=True, indent=2)
        return s

    def set_ploidy(self, haploid):
        if not haploid:
            return
        for k, v in self.items():
            if v.chr in haploid:
                v.ploidy = 1

    def get_info(self, tredName):
        tr = self.get(tredName)
        info = "END={};MOTIF={};NS=1;REF={};CR={};IH={};RL={};VT=STR".format(
            tr.repeat_end,
            tr.repeat,
            tr.ref_copy,
            tr.cutoff_risk,
            tr.inheritance,
            tr.ref_copy * len(tr.repeat),
        )
        return tr.chr, tr.repeat_start, tr.ref_copy, tr.repeat, info


class TRED(object):
    def __init__(self, name, row, ref=REF):

        self.row = row
        self.name = name
        self.repeat = row["repeat"]
        self.motif = row["motif"]
        repeat_location_field = "repeat_location"
        if ref != REF:
            repeat_location_field += "." + ref.split("_")[0]
        repeat_location = row[repeat_location_field]
        if "_nochr" in ref:  # Some reference version do not have chr
            repeat_location = repeat_location.replace("chr", "")
        self.chr, repeat_location = repeat_location.split(":")
        repeat_start, repeat_end = repeat_location.split("-")
        self.repeat_start = int(repeat_start)
        self.repeat_end = int(repeat_end)
        self.ref_copy = (self.repeat_end - self.repeat_start + 1) / len(self.repeat)
        self.prefix = row["prefix"]
        self.suffix = row["suffix"]
        self.cutoff_prerisk = row["cutoff_prerisk"]
        self.cutoff_risk = row["cutoff_risk"]
        self.inheritance = row["inheritance"]
        self.is_xlinked = self.inheritance[0] == "X"
        self.is_recessive = self.inheritance[-1] == "R"
        self.is_expansion = row["mutation_nature"] == "increase"
        self.ploidy = 2

    def __repr__(self):
        return "{} inheritance={} id={}_{}_{}".format(
            self.name, self.inheritance, self.chr, self.repeat_start, self.repeat
        )

    def __str__(self):
        return ";".join(
            str(x)
            for x in (
                self.name,
                self.repeat,
                self.chr,
                self.repeat_start,
                self.repeat_end,
                self.prefix,
                self.suffix,
            )
        )


class STRLine(object):
    def __init__(self, line):
        args = line.split()
        self.seqid = args[0]
        self.start = int(args[1])
        self.end = int(args[2])
        self.period = int(args[3])
        self.copynum = args[4]
        self.consensusSize = int(args[5])
        self.pctmatch = int(args[6])
        self.pctindel = int(args[7])
        self.score = args[8]
        self.A = args[9]
        self.C = args[10]
        self.G = args[11]
        self.T = args[12]
        self.entropy = float(args[13])
        self.motif = args[14]
        assert self.period == len(self.motif)
        self.name = args[15] if len(args) > 15 else None

    def __str__(self):
        fields = [
            self.seqid,
            self.start,
            self.end,
            self.period,
            self.copynum,
            self.consensusSize,
            self.pctmatch,
            self.pctindel,
            self.score,
            self.A,
            self.C,
            self.G,
            self.T,
            "{0:.2f}".format(self.entropy),
            self.motif,
        ]
        if self.name is not None:
            fields += [self.name]
        return "\t".join(str(x) for x in fields)

    @property
    def longname(self):
        return "_".join(str(x) for x in (self.seqid, self.start, self.motif))

    def is_valid(self, maxperiod=6, maxlength=READLEN, minscore=MINSCORE):
        return (
            1 <= self.period <= maxperiod
            and (self.end - self.start + 1) <= maxlength
            and self.score >= minscore
        )

    def calc_entropy(self):
        total = self.A + self.C + self.G + self.T
        if total == 0:  # Perhaps they are all Ns - might crash in lobstrindex()
            return 0
        fractions = [x * 1.0 / total for x in [self.A, self.C, self.G, self.T]]
        entropy = sum([-1.0 * x * log(x, 2) for x in fractions if x != 0])
        return entropy

    def iter_exact_str(self, genome):
        pat = re.compile("(({0}){{2,}})".format(self.motif))
        start = self.start
        s = genome[self.seqid][self.start - 1 : self.end].upper()
        for m in re.finditer(pat, s):
            self.start = start + m.start()
            length = m.end() - m.start()
            subseq = m.group(0)
            assert length % self.period == 0
            assert subseq.startswith(self.motif)

            self.end = self.start - 1 + length
            self.copynum = length / self.period
            self.pctmatch = 100
            self.pctindel = 0
            self.score = 2 * length
            self.fix_counts(subseq)
            yield self

    def fix_counts(self, subseq):
        length = int(ceil(self.period * float(self.copynum)))
        # Sanity check for length, otherwise lobSTR misses units
        self.end = max(self.end, self.start + length - 1)
        self.A = subseq.count("A")
        self.C = subseq.count("C")
        self.G = subseq.count("G")
        self.T = subseq.count("T")
        self.entropy = self.calc_entropy()


class STRFile(LineFile):
    def __init__(self, lobstr_home, db="hg38"):
        filename = op.join(lobstr_home, "{0}/index.info".format(db))
        super(STRFile, self).__init__(filename)
        fp = open(filename)
        for row in fp:
            self.append(STRLine(row))

    @property
    def ids(self):
        return [s.longname for s in self]

    @property
    def register(self):
        return dict(((s.seqid, s.start), s.name) for s in self)


class LobSTRvcf(dict):
    def __init__(self, columnidsfile="STR.ids"):
        super(LobSTRvcf, self).__init__()
        self.samplekey = None
        self.evidence = {}  # name: (supporting reads, stutter reads)
        if columnidsfile:
            fp = open(columnidsfile)
            self.columns = [x.strip() for x in fp]
            logging.debug(
                "A total of {} markers imported from `{}`".format(
                    len(self.columns), columnidsfile
                )
            )

    def parse(self, filename, filtered=True, cleanup=False):
        self.samplekey = op.basename(filename).split(".")[0]
        logging.debug("Parse `{}` (filtered={})".format(filename, filtered))
        fp = must_open(filename)
        reader = vcf.Reader(fp)
        for record in reader:
            if filtered and record.FILTER:
                continue
            info = record.INFO
            ref = float(info["REF"])
            rpa = info.get("RPA", ref)
            motif = info["MOTIF"]
            name = "_".join(str(x) for x in (record.CHROM, record.POS, motif))
            for sample in record.samples:
                gt = sample["GT"]
                if filtered and sample["FT"] != "PASS":
                    continue
                if gt == "0/0":
                    alleles = (ref, ref)
                elif gt in ("0/1", "1/0"):
                    alleles = (ref, rpa[0])
                elif gt == "1/1":
                    alleles = (rpa[0], rpa[0])
                elif gt == "1/2":
                    alleles = rpa
                try:
                    self[name] = ",".join(str(int(x)) for x in sorted(alleles))
                except:
                    self[name] = "-,-"

                # Collect supporting read evidence
                motif_length = len(motif)
                adjusted_alleles = [(x - ref) * motif_length for x in alleles]
                support = stutters = 0
                allreads = sample["ALLREADS"]
                for r in allreads.split(";"):
                    k, v = r.split("|")
                    k, v = int(k), int(v)
                    min_dist = min([abs(k - x) for x in adjusted_alleles])
                    if motif_length * 0.5 < min_dist < motif_length * 1.5:
                        stutters += v
                    support += v
                self.evidence[name] = "{},{}".format(stutters, support)

        if cleanup:
            sh("rm -f {}".format(op.basename(filename)))

    @property
    def csvline(self):
        return ",".join([self.get(c, "-1,-1") for c in self.columns])

    @property
    def evline(self):
        return ",".join([self.evidence.get(c, "-1,-1") for c in self.columns])


def main():

    actions = (
        # Compile population data - pipeline: compilevcf->mergecsv->meta->data->mask
        ("bin", "convert tsv to binary format"),
        ("filtervcf", "filter lobSTR VCF"),
        ("compilevcf", "compile vcf results into master spreadsheet"),
        ("mergecsv", "combine csv into binary array"),
        ("meta", "compute allele frequencies and write to meta"),
        ("data", "filter data based on the meta calls"),
        ("mask", "compute P-values based on meta calls and data"),
        ("treds", "compile allele_frequency for TRED results"),
        # lobSTR related
        ("lobstrindex", "make lobSTR index"),
        ("batchlobstr", "run batch lobSTR"),
        ("lobstr", "run lobSTR on a big BAM"),
        ("locus", "extract selected locus and run lobSTR"),
        ("stutter", "extract info from lobSTR vcf file"),
        # Specific markers
        ("liftover", "liftOver CODIS/Y-STR markers"),
        ("trf", "run TRF on FASTA files"),
        ("ystr", "print out Y-STR info given VCF"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def treds(args):
    """
    %prog treds hli.tred.tsv

    Compile allele_frequency for TREDs results. Write data.tsv, meta.tsv and
    mask.tsv in one go.
    """
    from jcvi.apps.base import datafile

    p = OptionParser(treds.__doc__)
    p.add_option(
        "--csv", default=False, action="store_true", help="Also write `meta.csv`"
    )
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (tredresults,) = args
    df = pd.read_csv(tredresults, sep="\t")

    tredsfile = datafile("TREDs.meta.csv")
    tf = pd.read_csv(tredsfile)

    tds = list(tf["abbreviation"])
    ids = list(tf["id"])
    tags = ["SampleKey"]
    final_columns = ["SampleKey"]
    afs = []
    for td, id in zip(tds, ids):
        tag1 = "{}.1".format(td)
        tag2 = "{}.2".format(td)
        if tag2 not in df:
            afs.append("{}")
            continue
        tags.append(tag2)
        final_columns.append(id)
        a = np.array(list(df[tag1]) + list(df[tag2]))
        counts = alleles_to_counts(a)
        af = counts_to_af(counts)
        afs.append(af)

    tf["allele_frequency"] = afs

    metafile = "TREDs_{}_SEARCH.meta.tsv".format(timestamp())
    tf.to_csv(metafile, sep="\t", index=False)
    logging.debug("File `{}` written.".format(metafile))
    if opts.csv:
        metacsvfile = metafile.rsplit(".", 1)[0] + ".csv"
        tf.to_csv(metacsvfile, index=False)
        logging.debug("File `{}` written.".format(metacsvfile))

    pp = df[tags]
    pp.columns = final_columns
    datafile = "TREDs_{}_SEARCH.data.tsv".format(timestamp())
    pp.to_csv(datafile, sep="\t", index=False)
    logging.debug("File `{}` written.".format(datafile))

    mask([datafile, metafile])


def stutter(args):
    """
    %prog stutter a.vcf.gz

    Extract info from lobSTR vcf file. Generates a file that has the following
    fields:

    CHR, POS, MOTIF, RL, ALLREADS, Q
    """
    p = OptionParser(stutter.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (vcf,) = args
    pf = op.basename(vcf).split(".")[0]
    execid, sampleid = pf.split("_")

    C = "vcftools --remove-filtered-all --min-meanDP 10"
    C += " --gzvcf {} --out {}".format(vcf, pf)
    C += " --indv {}".format(sampleid)

    info = pf + ".INFO"
    if need_update(vcf, info):
        cmd = C + " --get-INFO MOTIF --get-INFO RL"
        sh(cmd)

    allreads = pf + ".ALLREADS.FORMAT"
    if need_update(vcf, allreads):
        cmd = C + " --extract-FORMAT-info ALLREADS"
        sh(cmd)

    q = pf + ".Q.FORMAT"
    if need_update(vcf, q):
        cmd = C + " --extract-FORMAT-info Q"
        sh(cmd)

    outfile = pf + ".STUTTER"
    if need_update((info, allreads, q), outfile):
        cmd = "cut -f1,2,5,6 {}".format(info)
        cmd += r" | sed -e 's/\t/_/g'"
        cmd += " | paste - {} {}".format(allreads, q)
        cmd += " | cut -f1,4,7"
        sh(cmd, outfile=outfile)


def write_filtered(vcffile, lhome, store=None):
    if vcffile.startswith("s3://"):
        vcffile = pull_from_s3(vcffile)

    filteredvcf = op.basename(vcffile).replace(".vcf", ".filtered.vcf")
    cmd = "python {}/scripts/lobSTR_filter_vcf.py".format(lhome)
    cmd += " --vcf {}".format(vcffile)
    cmd += " --loc-cov 5 --loc-log-score 0.8"
    # cmd += " --loc-call-rate 0.8 --loc-max-ref-length 80"
    # cmd += " --call-cov 5 --call-log-score 0.8 --call-dist-end 20"
    sh(cmd, outfile=filteredvcf)

    if store:
        push_to_s3(store, filteredvcf)

    return filteredvcf


def run_filter(arg):
    vcffile, lhome, store = arg
    filteredvcf = vcffile.replace(".vcf", ".filtered.vcf")
    try:
        if vcffile.startswith("s3://"):
            if check_exists_s3(filteredvcf):
                logging.debug("{} exists. Skipped.".format(filteredvcf))
            else:
                write_filtered(vcffile, lhome, store=store)
                logging.debug("{} written and uploaded.".format(filteredvcf))
        else:
            if need_update(vcffile, filteredvcf):
                write_filtered(vcffile, lhome, store=None)
    except Exception as e:
        logging.debug("Thread failed! Error: {}".format(e))


def filtervcf(args):
    """
    %prog filtervcf NA12878.hg38.vcf.gz

    Filter lobSTR VCF using script shipped in lobSTR. Input file can be a list
    of vcf files.
    """
    p = OptionParser(filtervcf.__doc__)
    p.set_home("lobstr", default="/mnt/software/lobSTR")
    p.set_aws_opts(store="hli-mv-data-science/htang/str")
    p.set_cpus()
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (samples,) = args
    lhome = opts.lobstr_home
    store = opts.output_path

    if samples.endswith((".vcf", ".vcf.gz")):
        vcffiles = [samples]
    else:
        vcffiles = [x.strip() for x in must_open(samples)]

    vcffiles = [x for x in vcffiles if ".filtered." not in x]

    run_args = [(x, lhome, x.startswith("s3://") and store) for x in vcffiles]
    cpus = min(opts.cpus, len(run_args))
    p = Pool(processes=cpus)
    for _ in p.map_async(run_filter, run_args).get():
        continue


def write_meta(af_file, gene_map, blacklist, filename="meta.tsv"):
    fp = open(af_file)
    fw = open(filename, "w")
    header = "id title gene_name variant_type motif allele_frequency".replace(" ", "\t")
    print(header, file=fw)
    variant_type = "short tandem repeats"
    title = "Short tandem repeats ({})n"
    for row in fp:
        locus, af, remove = row.split()
        if remove == "MISSING":
            continue
        if locus in blacklist:
            continue

        seqid, pos, motif = locus.split("_")
        gene_name = gene_map.get((seqid, pos), "")
        print(
            "\t".join((locus, title.format(motif), gene_name, variant_type, motif, af)),
            file=fw,
        )
    fw.close()
    logging.debug("Write meta file to `{}`".format(filename))


def read_treds(tredsfile=datafile("TREDs.meta.csv")):
    if tredsfile.endswith(".csv"):
        df = pd.read_csv(tredsfile)
        treds = set(df["id"])
    else:
        df = pd.read_csv(tredsfile, sep="\t")
        treds = set(df["abbreviation"])

    logging.debug("Loaded {} treds from `{}`".format(len(treds), tredsfile))
    return treds, df


def meta(args):
    """
    %prog meta data.bin samples STR.ids STR-exons.wo.bed

    Compute allele frequencies and prune sites based on missingness.

    Filter subset of loci that satisfy:
    1. no redundancy (unique chr:pos)
    2. variable (n_alleles > 1)
    3. low level of missing data (>= 50% autosomal + X, > 25% for Y)

    Write meta file with the following infor:
    1. id
    2. title
    3. gene_name
    4. variant_type
    5. motif
    6. allele_frequency

    `STR-exons.wo.bed` can be generated like this:
    $ tail -n 694105 /mnt/software/lobSTR/hg38/index.tab | cut -f1-3 > all-STR.bed
    $ intersectBed -a all-STR.bed -b all-exons.bed -wo > STR-exons.wo.bed
    """
    p = OptionParser(meta.__doc__)
    p.add_option(
        "--cutoff",
        default=0.5,
        type="float",
        help="Percent observed required (chrY half cutoff)",
    )
    p.set_cpus()
    opts, args = p.parse_args(args)

    if len(args) != 4:
        sys.exit(not p.print_help())

    binfile, sampleids, strids, wobed = args
    cutoff = opts.cutoff

    af_file = "allele_freq"
    if need_update(binfile, af_file):
        df, m, samples, loci = read_binfile(binfile, sampleids, strids)
        nalleles = len(samples)
        fw = must_open(af_file, "w")
        for i, locus in enumerate(loci):
            a = m[:, i]
            counts = alleles_to_counts(a)
            af = counts_to_af(counts)
            seqid = locus.split("_")[0]
            remove = counts_filter(counts, nalleles, seqid, cutoff=cutoff)
            print("\t".join((locus, af, remove)), file=fw)
        fw.close()

    logging.debug("Load gene intersections from `{}`".format(wobed))
    fp = open(wobed)
    gene_map = defaultdict(set)
    for row in fp:
        chr1, start1, end1, chr2, start2, end2, name, ov = row.split()
        gene_map[(chr1, start1)] |= set(name.split(","))
    for k, v in gene_map.items():
        non_enst = sorted(x for x in v if not x.startswith("ENST"))
        # enst = sorted(x.rsplit(".", 1)[0] for x in v if x.startswith("ENST"))
        gene_map[k] = ",".join(non_enst)

    TREDS, df = read_treds()

    metafile = "STRs_{}_SEARCH.meta.tsv".format(timestamp())
    write_meta(af_file, gene_map, TREDS, filename=metafile)
    logging.debug("File `{}` written.".format(metafile))


def alleles_to_counts(a):
    # xa = a / 1000
    xb = a % 1000
    counts = Counter()
    # counts.update(xa)
    counts.update(xb)
    del counts[-1]
    del counts[999]
    return counts


def counts_to_af(counts):
    return "{" + ",".join("{}:{}".format(k, v) for k, v in sorted(counts.items())) + "}"


def af_to_counts(af):
    countst = [x for x in af.strip("{}").split(",") if x]
    countsd = {}
    for x in countst:
        a, b = x.split(":")
        countsd[int(a)] = int(b)
    return countsd


def bin(args):
    """
    %prog bin data.tsv

    Conver tsv to binary format.
    """
    p = OptionParser(bin.__doc__)
    p.add_option("--dtype", choices=("float32", "int32"), help="dtype of the matrix")
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (tsvfile,) = args
    dtype = opts.dtype
    if dtype is None:  # Guess
        dtype = np.int32 if "data" in tsvfile else np.float32
    else:
        dtype = np.int32 if dtype == "int32" else np.float32

    print("dtype: {}".format(dtype), file=sys.stderr)
    fp = open(tsvfile)
    next(fp)
    arrays = []
    for i, row in enumerate(fp):
        a = np.fromstring(row, sep="\t", dtype=dtype)
        a = a[1:]
        arrays.append(a)
        print(i, a, file=sys.stderr)

    print("Merging", file=sys.stderr)
    b = np.concatenate(arrays)
    print("Binary shape: {}".format(b.shape), file=sys.stderr)
    binfile = tsvfile.rsplit(".", 1)[0] + ".bin"
    b.tofile(binfile)


def counts_to_percentile(counts):
    percentile = {}
    s = 0
    for k, v in sorted(counts.items(), reverse=True):
        s += v
        percentile[k] = s
    for k, v in percentile.items():
        v = "{:.6f}".format(v * 1.0 / s)
        percentile[k] = v
    return percentile


def convert_to_percentile(arg):
    i, a, percentile = arg
    pp = np.array([percentile.get(x, "1.000000") for x in a], dtype="S8")
    if i % 1000 == 0:
        print(i, file=sys.stderr)
        print(a, file=sys.stderr)
        print(pp, file=sys.stderr)
    return i, pp


def write_csv(csvfile, m, index, columns, sep="\t", index_label="SampleKey"):
    fw = open(csvfile, "w")
    print(sep.join([index_label] + columns), file=fw)
    for i, a in enumerate(m):
        print(index[i] + sep + sep.join(str(x) for x in a), file=fw)
    fw.close()


def read_meta(metafile):
    df = pd.read_csv(metafile, sep="\t")
    final_columns = []
    percentiles = {}
    for i, row in df.iterrows():
        id = row["id"]
        final_columns.append(id)
        counts = row["allele_frequency"]
        countsd = af_to_counts(counts)
        percentile = counts_to_percentile(countsd)
        percentiles[id] = percentile
    return final_columns, percentiles


def write_mask(cpus, samples, final_columns, run_args, filename="mask.tsv"):
    p = Pool(processes=cpus)
    res = []
    r = p.map_async(convert_to_percentile, run_args, callback=res.append)
    r.wait()
    res.sort()

    if len(res) == 1:  # sometimes res end up with one more nest
        (res,) = res

    # Write mask (P-value) matrix
    ii, pvalues = zip(*res)
    m = np.vstack(pvalues).T
    write_csv(filename, m, samples, final_columns)


def data(args):
    """
    %prog data data.bin samples.ids STR.ids meta.tsv

    Make data.tsv based on meta.tsv.
    """
    p = OptionParser(data.__doc__)
    p.add_option(
        "--notsv", default=False, action="store_true", help="Do not write data.tsv"
    )
    opts, args = p.parse_args(args)

    if len(args) != 4:
        sys.exit(not p.print_help())

    databin, sampleids, strids, metafile = args
    final_columns, percentiles = read_meta(metafile)
    df, m, samples, loci = read_binfile(databin, sampleids, strids)

    # Clean the data
    m %= 1000  # Get the larger of the two alleles
    m[m == 999] = -1  # Missing data

    final = set(final_columns)
    remove = []
    for i, locus in enumerate(loci):
        if locus not in final:
            remove.append(locus)
            continue

    pf = "STRs_{}_SEARCH".format(timestamp())
    filteredstrids = "{}.STR.ids".format(pf)
    fw = open(filteredstrids, "w")
    print("\n".join(final_columns), file=fw)
    fw.close()
    logging.debug(
        "Dropped {} columns; Retained {} columns (`{}`)".format(
            len(remove), len(final_columns), filteredstrids
        )
    )

    # Remove low-quality columns!
    df.drop(remove, inplace=True, axis=1)
    df.columns = final_columns

    filtered_bin = "{}.data.bin".format(pf)
    if need_update(databin, filtered_bin):
        m = df.as_matrix()
        m.tofile(filtered_bin)
        logging.debug("Filtered binary matrix written to `{}`".format(filtered_bin))

    # Write data output
    filtered_tsv = "{}.data.tsv".format(pf)
    if not opts.notsv and need_update(databin, filtered_tsv):
        df.to_csv(filtered_tsv, sep="\t", index_label="SampleKey")


def mask(args):
    """
    %prog mask data.bin samples.ids STR.ids meta.tsv

    OR

    %prog mask data.tsv meta.tsv

    Compute P-values based on meta and data. The `data.bin` should be the matrix
    containing filtered loci and the output mask.tsv will have the same
    dimension.
    """
    p = OptionParser(mask.__doc__)
    opts, args = p.parse_args(args)

    if len(args) not in (2, 4):
        sys.exit(not p.print_help())

    if len(args) == 4:
        databin, sampleids, strids, metafile = args
        df, m, samples, loci = read_binfile(databin, sampleids, strids)
        mode = "STRs"
    elif len(args) == 2:
        databin, metafile = args
        df = pd.read_csv(databin, sep="\t", index_col=0)
        m = df.as_matrix()
        samples = df.index
        loci = list(df.columns)
        mode = "TREDs"

    pf = "{}_{}_SEARCH".format(mode, timestamp())
    final_columns, percentiles = read_meta(metafile)

    maskfile = pf + ".mask.tsv"
    run_args = []
    for i, locus in enumerate(loci):
        a = m[:, i]
        percentile = percentiles[locus]
        run_args.append((i, a, percentile))

    if mode == "TREDs" or need_update(databin, maskfile):
        cpus = min(8, len(run_args))
        write_mask(cpus, samples, final_columns, run_args, filename=maskfile)
        logging.debug("File `{}` written.".format(maskfile))


def counts_filter(countsd, nalleles, seqid, cutoff=0.5):
    cutoff *= 100
    # Check for missingness
    observed = sum(countsd.values())
    observed_pct = observed * 100 / nalleles
    if observed_pct < cutoff:
        if not (seqid == "chrY" and observed_pct >= cutoff / 2):
            return "MISSING"

    # Check for variability
    if len(countsd) < 2:
        return "INVARIANT"

    return "PASS"


def read_binfile(binfile, sampleids, strids, dtype=np.int32):
    m = np.fromfile(binfile, dtype=dtype)
    samples = [x.strip() for x in open(sampleids)]
    loci = [x.strip() for x in open(strids)]
    nsamples, nloci = len(samples), len(loci)
    print("{} x {} entries imported".format(nsamples, nloci), file=sys.stderr)

    m.resize(nsamples, nloci)
    df = pd.DataFrame(m, index=samples, columns=loci)
    return df, m, samples, loci


def mergecsv(args):
    """
    %prog mergecsv *.csv

    Combine CSV into binary array.
    """
    p = OptionParser(mergecsv.__doc__)
    opts, args = p.parse_args(args)

    if len(args) < 1:
        sys.exit(not p.print_help())

    csvfiles = args
    arrays = []
    samplekeys = []
    for csvfile in csvfiles:
        samplekey = op.basename(csvfile).split(".")[0]
        a = np.fromfile(csvfile, sep=",", dtype=np.int32)
        x1 = a[::2]
        x2 = a[1::2]
        a = x1 * 1000 + x2
        a[a < 0] = -1
        arrays.append(a)
        samplekeys.append(samplekey)
        print(samplekey, a, file=sys.stderr)
    print("Merging", file=sys.stderr)
    b = np.concatenate(arrays)
    b.tofile("data.bin")

    fw = open("samples", "w")
    print("\n".join(samplekeys), file=fw)
    fw.close()


def write_csv_ev(filename, filtered, cleanup, store=None):
    lv = LobSTRvcf()
    lv.parse(filename, filtered=filtered, cleanup=cleanup)
    csvfile = op.basename(filename) + ".csv"
    evfile = op.basename(filename) + ".ev"

    fw = open(csvfile, "w")
    print(lv.csvline, file=fw)
    fw.close()

    fw = open(evfile, "w")
    print(lv.evline, file=fw)
    fw.close()

    # Save to s3
    if store:
        push_to_s3(store, csvfile)
        push_to_s3(store, evfile)


def run_compile(arg):
    filename, filtered, cleanup, store = arg
    csvfile = filename + ".csv"
    try:
        if filename.startswith("s3://"):
            if check_exists_s3(csvfile):
                logging.debug("{} exists. Skipped.".format(csvfile))
            else:
                write_csv_ev(filename, filtered, cleanup, store=store)
                logging.debug("{} written and uploaded.".format(csvfile))
        else:
            if need_update(filename, csvfile):
                write_csv_ev(filename, filtered, cleanup, store=None)
    except Exception as e:
        logging.debug("Thread failed! Error: {}".format(e))


def compilevcf(args):
    """
    %prog compilevcf samples.csv

    Compile vcf results into master spreadsheet.
    """
    p = OptionParser(compilevcf.__doc__)
    p.add_option("--db", default="hg38", help="Use these lobSTR db")
    p.add_option(
        "--nofilter",
        default=False,
        action="store_true",
        help="Do not filter the variants",
    )
    p.set_home("lobstr")
    p.set_cpus()
    p.set_aws_opts(store="hli-mv-data-science/htang/str-data")
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (samples,) = args
    workdir = opts.workdir
    store = opts.output_path
    cleanup = not opts.nocleanup
    filtered = not opts.nofilter
    dbs = opts.db.split(",")
    cwd = os.getcwd()
    mkdir(workdir)
    os.chdir(workdir)
    samples = op.join(cwd, samples)

    stridsfile = "STR.ids"
    if samples.endswith((".vcf", ".vcf.gz")):
        vcffiles = [samples]
    else:
        vcffiles = [x.strip() for x in must_open(samples)]
    if not op.exists(stridsfile):
        ids = []
        for db in dbs:
            ids.extend(STRFile(opts.lobstr_home, db=db).ids)
        uids = uniqify(ids)
        logging.debug("Combined: {} Unique: {}".format(len(ids), len(uids)))

        fw = open(stridsfile, "w")
        print("\n".join(uids), file=fw)
        fw.close()

    run_args = [(x, filtered, cleanup, store) for x in vcffiles]
    cpus = min(opts.cpus, len(run_args))
    p = Pool(processes=cpus)
    for _ in p.map_async(run_compile, run_args).get():
        continue


def build_ysearch_link(r, ban=["DYS520", "DYS413a", "DYS413b"]):
    template = "http://www.ysearch.org/search_search.asp?fail=2&uid=&freeentry=true&"
    markers = []
    for i, marker in zip(YSEARCH_LL, YSEARCH_HAPLOTYPE):
        z = r.get(marker, "null")
        if "a/b" in marker or marker in ban:
            z = "null"
        m = "{0}={1}".format(i, z)
        markers.append(m)
    print(template + "&".join(markers))


def build_yhrd_link(r, panel, ban=["DYS385"]):
    L = []
    for marker in panel:
        z = r.get(marker, "--")
        if marker in ban:
            z = "--"
        L.append(z)
    print(" ".join(str(x) for x in L))


def ystr(args):
    """
    %prog ystr chrY.vcf

    Print out Y-STR info given VCF. Marker name extracted from tabfile.
    """
    from jcvi.utils.table import write_csv

    p = OptionParser(ystr.__doc__)
    p.set_home("lobstr")
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (vcffile,) = args
    si = STRFile(opts.lobstr_home, db="hg38-named")
    register = si.register

    header = "Marker|Reads|Ref|Genotype|Motif".split("|")
    contents = []
    fp = must_open(vcffile)
    reader = vcf.Reader(fp)
    simple_register = {}
    for record in reader:
        name = register[(record.CHROM, record.POS)]
        info = record.INFO
        ref = int(float(info["REF"]))
        rpa = info.get("RPA", ref)
        if isinstance(rpa, list):
            rpa = "|".join(str(int(float(x))) for x in rpa)
        ru = info["RU"]
        simple_register[name] = rpa
        for sample in record.samples:
            contents.append((name, sample["ALLREADS"], ref, rpa, ru))

    # Multi-part markers
    a, b, c = "DYS389I", "DYS389B.1", "DYS389B"
    if a in simple_register and b in simple_register:
        simple_register[c] = int(simple_register[a]) + int(simple_register[b])

    # Multi-copy markers
    mm = ["DYS385", "DYS413", "YCAII"]
    for m in mm:
        ma, mb = m + "a", m + "b"
        if ma not in simple_register or mb not in simple_register:
            simple_register[ma] = simple_register[mb] = None
            del simple_register[ma]
            del simple_register[mb]
            continue
        if simple_register[ma] > simple_register[mb]:
            simple_register[ma], simple_register[mb] = (
                simple_register[mb],
                simple_register[ma],
            )

    write_csv(header, contents, sep=" ")
    print("[YSEARCH]")
    build_ysearch_link(simple_register)
    print("[YFILER]")
    build_yhrd_link(simple_register, panel=YHRD_YFILER)
    print("[YFILERPLUS]")
    build_yhrd_link(simple_register, panel=YHRD_YFILERPLUS)
    print("[YSTR-ALL]")
    build_yhrd_link(simple_register, panel=USYSTR_ALL)


def get_motif(s, motif_length):
    sl = len(s)
    kmers = set()
    # Get all kmers
    for i in range(sl - motif_length):
        ss = s[i : i + motif_length]
        kmers.add(ss)

    kmer_counts = []
    for kmer in kmers:
        kmer_counts.append((s.count(kmer), -s.index(kmer), kmer))

    return sorted(kmer_counts, reverse=True)[0][-1]


def liftover(args):
    """
    %prog liftover lobstr_v3.0.2_hg38_ref.bed hg38.upper.fa

    LiftOver CODIS/Y-STR markers.
    """
    p = OptionParser(liftover.__doc__)
    p.add_option(
        "--checkvalid",
        default=False,
        action="store_true",
        help="Check minscore, period and length",
    )
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    refbed, fastafile = args
    genome = pyfasta.Fasta(fastafile)
    edits = []
    fp = open(refbed)
    for i, row in enumerate(fp):
        s = STRLine(row)
        seq = genome[s.seqid][s.start - 1 : s.end].upper()
        s.motif = get_motif(seq, len(s.motif))
        s.fix_counts(seq)
        if opts.checkvalid and not s.is_valid():
            continue
        edits.append(s)
        if i % 10000 == 0:
            print(i, "lines read", file=sys.stderr)

    edits = natsorted(edits, key=lambda x: (x.seqid, x.start))
    for e in edits:
        print(str(e))


def trf(args):
    """
    %prog trf outdir

    Run TRF on FASTA files.
    """
    from jcvi.apps.base import iglob

    cparams = "1 1 2 80 5 200 2000"

    p = OptionParser(trf.__doc__)
    p.add_option("--mismatch", default=31, type="int", help="Mismatch and gap penalty")
    p.add_option(
        "--minscore", default=MINSCORE, type="int", help="Minimum score to report"
    )
    p.add_option("--period", default=6, type="int", help="Maximum period to report")
    p.add_option(
        "--lobstr",
        default=False,
        action="store_true",
        help="Generate output for lobSTR",
    )
    p.add_option(
        "--telomeres",
        default=False,
        action="store_true",
        help="Run telomere search: minscore=140 period=7",
    )
    p.add_option(
        "--centromeres",
        default=False,
        action="store_true",
        help="Run centromere search: {}".format(cparams),
    )
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (outdir,) = args
    minlength = opts.minscore / 2
    mm = MakeManager()
    if opts.telomeres:
        opts.minscore, opts.period = 140, 7

    params = "2 {0} {0} 80 10 {1} {2}".format(
        opts.mismatch, opts.minscore, opts.period
    ).split()
    if opts.centromeres:
        params = cparams.split()

    bedfiles = []
    for fastafile in natsorted(iglob(outdir, "*.fa,*.fasta")):
        pf = op.basename(fastafile).rsplit(".", 1)[0]
        # Commands starting with trf ignores errors
        cmd1 = "-trf {0} {1} -d -h".format(fastafile, " ".join(params))
        datfile = op.basename(fastafile) + "." + ".".join(params) + ".dat"
        bedfile = "{0}.trf.bed".format(pf)
        cmd2 = "cat {} | grep -v ^Parameters".format(datfile)
        if opts.lobstr:
            cmd2 += " | awk '($8 >= {} && $8 <= {})'".format(
                minlength, READLEN - minlength
            )
        else:
            cmd2 += " | awk '($8 >= 0)'"
        cmd2 += " | sed 's/ /\\t/g'"
        cmd2 += " | awk '{{print \"{0}\\t\" $0}}' > {1}".format(pf, bedfile)
        mm.add(fastafile, datfile, cmd1)
        mm.add(datfile, bedfile, cmd2)
        bedfiles.append(bedfile)

    bedfile = "trf.bed"
    cmd = "cat {0} > {1}".format(" ".join(natsorted(bedfiles)), bedfile)
    mm.add(bedfiles, bedfile, cmd)

    mm.write()


def batchlobstr(args):
    """
    %prog batchlobstr samples.csv

    Run lobSTR sequentially on list of samples. Each line contains:
    sample-name,s3-location
    """
    p = OptionParser(batchlobstr.__doc__)
    p.add_option("--sep", default=",", help="Separator for building commandline")
    p.set_home("lobstr", default="s3://hli-mv-data-science/htang/str-build/lobSTR/")
    p.set_aws_opts(store="hli-mv-data-science/htang/str-data")
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (samplesfile,) = args
    store = opts.output_path
    computed = ls_s3(store)
    fp = open(samplesfile)
    skipped = total = 0
    for row in fp:
        total += 1
        sample, s3file = row.strip().split(",")[:2]
        exec_id, sample_id = sample.split("_")
        bamfile = s3file.replace(".gz", "").replace(".vcf", ".bam")

        gzfile = sample + ".{0}.vcf.gz".format("hg38")
        if gzfile in computed:
            skipped += 1
            continue

        print(
            opts.sep.join(
                "python -m jcvi.variation.str lobstr".split()
                + [
                    "hg38",
                    "--input_bam_path",
                    bamfile,
                    "--output_path",
                    store,
                    "--sample_id",
                    sample_id,
                    "--workflow_execution_id",
                    exec_id,
                    "--lobstr_home",
                    opts.lobstr_home,
                    "--workdir",
                    opts.workdir,
                ]
            )
        )
    fp.close()
    logging.debug("Total skipped: {0}".format(percentage(skipped, total)))


def lobstr(args):
    """
    %prog lobstr lobstr_index1 lobstr_index2 ...

    Run lobSTR on a big BAM file. There can be multiple lobSTR indices. In
    addition, bamfile can be S3 location and --lobstr_home can be S3 location
    (e.g. s3://hli-mv-data-science/htang/str-build/lobSTR/)
    """
    p = OptionParser(lobstr.__doc__)
    p.add_option(
        "--haploid", default="chrY,chrM", help="Use haploid model for these chromosomes"
    )
    p.add_option("--chr", help="Run only this chromosome")
    p.add_option(
        "--simulation", default=False, action="store_true", help="Simulation mode"
    )
    p.set_home("lobstr", default="s3://hli-mv-data-science/htang/str-build/lobSTR/")
    p.set_cpus()
    p.set_aws_opts(store="hli-mv-data-science/htang/str-data")
    opts, args = p.parse_args(args)
    bamfile = opts.input_bam_path

    if len(args) < 1 or bamfile is None:
        sys.exit(not p.print_help())

    lbindices = args
    if opts.simulation:  # Simulation mode
        cmd, vcf_file = allelotype_on_chr(
            bamfile, "chr4", "/mnt/software/lobSTR/", "TREDs", haploid=opts.haploid
        )
        stats_file = vcf_file.rsplit(".", 1)[0] + ".allelotype.stats"
        results_dir = "lobstr_results"
        mkdir(results_dir)
        sh(cmd)
        sh("mv {} {}/ && rm {}".format(vcf_file, results_dir, stats_file))
        return

    s3mode = bamfile.startswith("s3")
    store = opts.output_path
    cleanup = not opts.nocleanup
    workdir = opts.workdir
    mkdir(workdir)
    os.chdir(workdir)

    lhome = opts.lobstr_home
    if lhome.startswith("s3://"):
        lhome = pull_from_s3(lhome, overwrite=False)

    exec_id, sample_id = opts.workflow_execution_id, opts.sample_id
    prefix = [x for x in (exec_id, sample_id) if x]
    if prefix:
        pf = "_".join(prefix)
    else:
        pf = bamfile.split("/")[-1].split(".")[0]

    if s3mode:
        gzfile = pf + ".{0}.vcf.gz".format(lbindices[-1])
        remotegzfile = "{0}/{1}".format(store, gzfile)
        if check_exists_s3(remotegzfile):
            logging.debug(
                "Object `{0}` exists. Computation skipped.".format(remotegzfile)
            )
            return
        localbamfile = pf + ".bam"
        localbaifile = localbamfile + ".bai"
        if op.exists(localbamfile):
            logging.debug("BAM file already downloaded.")
        else:
            pull_from_s3(bamfile, localbamfile)
        if op.exists(localbaifile):
            logging.debug("BAM index file already downloaded.")
        else:
            remotebaifile = bamfile + ".bai"
            if check_exists_s3(remotebaifile):
                pull_from_s3(remotebaifile, localbaifile)
            else:
                remotebaifile = bamfile.rsplit(".")[0] + ".bai"
                if check_exists_s3(remotebaifile):
                    pull_from_s3(remotebaifile, localbaifile)
                else:
                    logging.debug("BAM index cannot be found in S3!")
                    sh("samtools index {0}".format(localbamfile))
        bamfile = localbamfile

    chrs = [opts.chr] if opts.chr else (range(1, 23) + ["X", "Y"])
    for lbidx in lbindices:
        makefile = "makefile.{0}".format(lbidx)
        mm = MakeManager(filename=makefile)
        vcffiles = []
        for chr in chrs:
            cmd, vcffile = allelotype_on_chr(
                bamfile, chr, lhome, lbidx, haploid=opts.haploid
            )
            mm.add(bamfile, vcffile, cmd)
            filteredvcffile = vcffile.replace(".vcf", ".filtered.vcf")
            cmd = "python -m jcvi.variation.str filtervcf {}".format(vcffile)
            cmd += " --lobstr_home {}".format(lhome)
            mm.add(vcffile, filteredvcffile, cmd)
            vcffiles.append(filteredvcffile)

        gzfile = bamfile.split(".")[0] + ".{0}.vcf.gz".format(lbidx)
        cmd = "vcf-concat {0} | vcf-sort".format(" ".join(vcffiles))
        cmd += " | bgzip -c > {0}".format(gzfile)
        mm.add(vcffiles, gzfile, cmd)

        mm.run(cpus=opts.cpus)

        if s3mode:
            push_to_s3(store, gzfile)

    if cleanup:
        mm.clean()
        sh("rm -f {} {} *.bai *.stats".format(bamfile, mm.makefile))


def allelotype_on_chr(bamfile, chr, lhome, lbidx, haploid="chrY,chrM"):
    if "chr" not in chr.lower():
        chr = "chr" + chr
    outfile = "{0}.{1}".format(bamfile.split(".")[0], chr)
    cmd = "allelotype --command classify --bam {}".format(bamfile)
    cmd += " --noise_model {0}/models/illumina_v3.pcrfree".format(lhome)
    cmd += " --strinfo {0}/{1}/index.tab".format(lhome, lbidx)
    cmd += " --index-prefix {0}/{1}/lobSTR_".format(lhome, lbidx)
    cmd += " --chrom {0} --out {1}.{2}".format(chr, outfile, lbidx)
    cmd += " --max-diff-ref {0}".format(READLEN)
    cmd += " --realign"
    cmd += " --haploid {}".format(haploid)
    return cmd, ".".join((outfile, lbidx, "vcf"))


def locus(args):
    """
    %prog locus bamfile

    Extract selected locus from a list of TREDs for validation, and run lobSTR.
    """
    from jcvi.formats.sam import get_minibam

    # See `Format-lobSTR-database.ipynb` for a list of TREDs for validation
    INCLUDE = ["HD", "SBMA", "SCA1", "SCA2", "SCA8", "SCA17", "DM1", "DM2", "FXTAS"]
    db_choices = ("hg38", "hg19")

    p = OptionParser(locus.__doc__)
    p.add_option("--tred", choices=INCLUDE, help="TRED name")
    p.add_option("--ref", choices=db_choices, default="hg38", help="Reference genome")
    p.set_home("lobstr")
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (bamfile,) = args
    ref = opts.ref
    lhome = opts.lobstr_home
    tred = opts.tred

    tredsfile = datafile("TREDs.meta.csv")
    tf = pd.read_csv(tredsfile, index_col=0)
    row = tf.ix[tred]
    tag = "repeat_location"
    ldb = "TREDs"
    if ref == "hg19":
        tag += "." + ref
        ldb += "-" + ref
    seqid, start_end = row[tag].split(":")

    PAD = 1000
    start, end = start_end.split("-")
    start, end = int(start) - PAD, int(end) + PAD
    region = "{}:{}-{}".format(seqid, start, end)

    minibamfile = get_minibam(bamfile, region)
    c = seqid.replace("chr", "")
    cmd, vcf = allelotype_on_chr(minibamfile, c, lhome, ldb)
    sh(cmd)

    parser = LobSTRvcf(columnidsfile=None)
    parser.parse(vcf, filtered=False)
    items = parser.items()
    if not items:
        print("No entry found!", file=sys.stderr)
        return

    k, v = parser.items()[0]
    print("{} => {}".format(tred, v.replace(",", "/")), file=sys.stderr)


def lobstrindex(args):
    """
    %prog lobstrindex hg38.trf.bed hg38.upper.fa

    Make lobSTR index. Make sure the FASTA contain only upper case (so use
    fasta.format --upper to convert from UCSC fasta). The bed file is generated
    by str().
    """
    p = OptionParser(lobstrindex.__doc__)
    p.add_option(
        "--notreds",
        default=False,
        action="store_true",
        help="Remove TREDs from the bed file",
    )
    p.set_home("lobstr")
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    trfbed, fastafile = args
    pf = fastafile.split(".")[0]
    lhome = opts.lobstr_home
    mkdir(pf)

    if opts.notreds:
        newbedfile = trfbed + ".new"
        newbed = open(newbedfile, "w")
        fp = open(trfbed)
        retained = total = 0
        seen = set()
        for row in fp:
            r = STRLine(row)
            total += 1
            name = r.longname
            if name in seen:
                continue
            seen.add(name)
            print(r, file=newbed)
            retained += 1
        newbed.close()
        logging.debug("Retained: {0}".format(percentage(retained, total)))
    else:
        newbedfile = trfbed

    mm = MakeManager()
    cmd = "python {0}/scripts/lobstr_index.py".format(lhome)
    cmd += " --str {0} --ref {1} --out {2}".format(newbedfile, fastafile, pf)
    mm.add((newbedfile, fastafile), op.join(pf, "lobSTR_ref.fasta.rsa"), cmd)

    tabfile = "{0}/index.tab".format(pf)
    cmd = "python {0}/scripts/GetSTRInfo.py".format(lhome)
    cmd += " {0} {1} > {2}".format(newbedfile, fastafile, tabfile)
    mm.add((newbedfile, fastafile), tabfile, cmd)

    infofile = "{0}/index.info".format(pf)
    cmd = "cp {0} {1}".format(newbedfile, infofile)
    mm.add(trfbed, infofile, cmd)
    mm.write()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
From genomeCovergeBed results, initialize the count array, set cutoffs
and optimize against the truth, to determine the cutoff for incorporating
RNA-seq into annotation pipelines.
"""
import sys
import os.path as op
import numpy as np
import logging

from itertools import groupby

from jcvi.formats.sizes import Sizes
from jcvi.formats.base import BaseFile, must_open
from jcvi.apps.base import OptionParser, ActionDispatcher


class BinFile(BaseFile):
    """
    The binfile contains per base count, fastafile provides the coordinate
    system.
    """

    def __init__(self, binfile, dtype=np.uint8):
        super(BinFile, self).__init__(binfile)
        assert op.exists(
            binfile
        ), "Binary file `{0}` not found. Rerun depth.count().".format(binfile)
        self.dtype = dtype

    @property
    def array(self):
        binfile = self.filename
        return np.fromfile(binfile, dtype=self.dtype)

    @property
    def mmarray(self):
        binfile = self.filename
        return np.memmap(binfile, dtype=self.dtype, mode="r")


def main():

    actions = (
        ("count", "initialize the count array"),
        ("query", "query the count array to get depth at particular site"),
        ("merge", "merge several count arrays into one"),
        (
            "bed",
            "write bed files where the bases have at least certain depth",
        ),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def bed(args):
    """
    %prog bed binfile fastafile

    Write bed files where the bases have at least certain depth.
    """
    p = OptionParser(bed.__doc__)
    p.add_option(
        "-o",
        dest="output",
        default="stdout",
        help="Output file name",
    )
    p.add_option(
        "--cutoff",
        dest="cutoff",
        default=10,
        type="int",
        help="Minimum read depth to report intervals",
    )
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    binfile, fastafile = args
    fw = must_open(opts.output, "w")
    cutoff = opts.cutoff
    assert cutoff >= 0, "Need non-negative cutoff"

    b = BinFile(binfile)
    ar = b.array

    fastasize, sizes, offsets = get_offsets(fastafile)
    s = Sizes(fastafile)
    for ctg, ctglen in s.iter_sizes():
        offset = offsets[ctg]
        subarray = ar[offset : offset + ctglen]
        key = lambda x: x[1] >= cutoff
        for tf, array_elements in groupby(enumerate(subarray), key=key):
            array_elements = list(array_elements)
            if not tf:
                continue

            # 0-based system => 1-based system
            start = array_elements[0][0] + 1
            end = array_elements[-1][0] + 1

            mean_depth = sum([x[1] for x in array_elements]) / len(array_elements)
            mean_depth = int(mean_depth)

            name = "na"
            print(
                "\t".join(str(x) for x in (ctg, start - 1, end, name, mean_depth)),
                file=fw,
            )


def merge(args):
    """
    %prog merge *.bin merged.bin

    Merge several count arrays into one. Overflows will be capped at uint8_max
    (255).
    """
    p = OptionParser(merge.__doc__)
    opts, args = p.parse_args(args)

    if len(args) < 2:
        sys.exit(not p.print_help())

    binfiles = args[:-1]
    mergedbin = args[-1]
    if op.exists(mergedbin):
        logging.error("`{0}` file exists. Remove before proceed.".format(mergedbin))
        return

    b = BinFile(binfiles[0])
    ar = b.mmarray
    (fastasize,) = ar.shape
    logging.debug("Initialize array of uint16 with size {0}".format(fastasize))

    merged_ar = np.zeros(fastasize, dtype=np.uint16)
    for binfile in binfiles:
        b = BinFile(binfile)
        merged_ar += b.array

    logging.debug("Resetting the count max to 255.")
    merged_ar[merged_ar > 255] = 255

    logging.debug("Compact array back to uint8 with size {0}".format(fastasize))
    merged_ar = np.array(merged_ar, dtype=np.uint8)
    merged_ar.tofile(mergedbin)
    logging.debug("Merged array written to `{0}`".format(mergedbin))


def query(args):
    """
    %prog query binfile fastafile ctgID baseID

    Get the depth at a particular base.
    """
    p = OptionParser(query.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 4:
        sys.exit(not p.print_help())

    binfile, fastafile, ctgID, baseID = args
    b = BinFile(binfile)
    ar = b.mmarray

    fastasize, sizes, offsets = get_offsets(fastafile)
    oi = offsets[ctgID] + int(baseID) - 1
    print("\t".join((ctgID, baseID, str(ar[oi]))))


def update_array(ar, coveragefile, offsets):
    fp = open(coveragefile)
    logging.debug("Parse file `{0}`".format(coveragefile))
    for k, rows in groupby(fp, key=(lambda x: x.split()[0])):
        rows = list(rows)
        offset = offsets[k]
        ctglen = len(rows)

        if ctglen < 100000:
            sys.stdout.write(".")
        else:
            print(k, offset)

        # assert ctglen == sizes[k]
        for i, row in enumerate(rows):
            ctgID, baseID, count = row.split()
            oi = offset + i
            newcount = ar[oi] + int(count)
            if newcount > 255:
                newcount = 255

            ar[oi] = newcount


def get_offsets(fastafile):
    s = Sizes(fastafile)
    fastasize = s.totalsize
    sizes = s.mapping
    offsets = s.cumsizes_mapping
    return fastasize, sizes, offsets


def count(args):
    """
    %prog count t.coveragePerBase fastafile

    Serialize the genomeCoverage results. The coordinate system of the count array
    will be based on the fastafile.
    """
    p = OptionParser(count.__doc__)
    opts, args = p.parse_args(args)

    if len(args) != 2:
        sys.exit(not p.print_help())

    coveragefile, fastafile = args

    countsfile = coveragefile.split(".")[0] + ".bin"
    if op.exists(countsfile):
        logging.error("`{0}` file exists. Remove before proceed.".format(countsfile))
        return

    fastasize, sizes, offsets = get_offsets(fastafile)
    logging.debug("Initialize array of uint8 with size {0}".format(fastasize))
    ar = np.zeros(fastasize, dtype=np.uint8)

    update_array(ar, coveragefile, offsets)

    ar.tofile(countsfile)
    logging.debug("Array written to `{0}`".format(countsfile))


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: UTF-8 -*-

"""
Mask low complexity regions in the genome.
"""
import os.path as op
import sys

from jcvi.formats.fasta import Fasta
from jcvi.utils.cbook import depends, percentage
from jcvi.apps.base import OptionParser, ActionDispatcher, sh


@depends
def wm_mk_counts(infile=None, outfile=None):
    cmd = "windowmasker -in {0} -mk_counts".format(infile)
    cmd += " -out {0}".format(outfile)
    sh(cmd)


@depends
def wm_mk_masks(infile=None, outfile=None, genomefile=None):
    cmd = "windowmasker -in {0} -ustat {1}".format(genomefile, infile)
    cmd += " -outfmt fasta -dust T -out {0}".format(outfile)
    sh(cmd)


def hardmask(fastafile):
    cmd = "maskOutFa {0} hard {0}".format(fastafile)
    sh(cmd)


def main():

    actions = (
        ("mask", "use windowmasker to mask low-complexity bases"),
        ("summary", "report the number of bases and sequences masked"),
    )
    p = ActionDispatcher(actions)
    p.dispatch(globals())


def summary(args):
    """
    %prog summary fastafile

    Report the number of bases and sequences masked.
    """
    p = OptionParser(summary.__doc__)

    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (fastafile,) = args
    f = Fasta(fastafile, index=False)

    halfmaskedseqs = set()
    allmasked = 0
    allbases = 0
    cutoff = 50
    for key, seq in f.iteritems():
        masked = 0
        for base in seq:
            if base not in "AGCT":
                masked += 1
        seqlen = len(seq)
        if masked * 100.0 / seqlen > cutoff:
            halfmaskedseqs.add(key)
        allmasked += masked
        allbases += seqlen

    seqnum = len(f)
    maskedseqnum = len(halfmaskedseqs)

    print(
        "Total masked bases: {0}".format(percentage(allmasked, allbases)),
        file=sys.stderr,
    )
    print(
        "Total masked sequences (contain > {0}% masked): {1}".format(
            cutoff, percentage(maskedseqnum, seqnum)
        ),
        file=sys.stderr,
    )


def mask(args):
    """
    %prog mask fastafile

    This script pipelines the windowmasker in NCBI BLAST+.  Masked fasta file
    will have an appended suffix of .mask with all the low-complexity bases masked
    (default to lower case, set --hard for hardmasking).
    """
    p = OptionParser(mask.__doc__)
    p.add_option(
        "--hard",
        dest="hard",
        default=False,
        action="store_true",
        help="Hard mask the low-complexity bases",
    )
    opts, args = p.parse_args(args)

    if len(args) != 1:
        sys.exit(not p.print_help())

    (genomefile,) = args

    # entire pipeline
    countsfile = genomefile + ".counts"
    wm_mk_counts(infile=genomefile, outfile=countsfile)

    maskedfastafile = "%s.masked%s" % op.splitext(genomefile)
    wm_mk_masks(infile=countsfile, outfile=maskedfastafile, genomefile=genomefile)

    if opts.hard:
        hardmask(maskedfastafile)


if __name__ == "__main__":
    main()
